from django import template

from ..models import EveType

register = template.Library()


@register.inclusion_tag("structures/templatetags/detail_title.html")
def detail_title(structure):
    """Render HTML for detail box title."""
    return {"image_url": structure.eve_type.render_url(256)}


@register.inclusion_tag("structures/templatetags/list_title.html")
def list_title(title: str):
    """Render HTML for list title."""
    return {"title": title}


@register.inclusion_tag("structures/templatetags/list_item.html")
def list_item(title: str, value="", eve_type=None, is_muted=False, url=None):
    """Render HTML for list item."""
    return {
        "title": title,
        "value": value,
        "eve_type": eve_type,
        "is_muted": is_muted,
        "url": url,
    }


@register.inclusion_tag("structures/templatetags/list_asset.html")
def list_asset(eve_type: EveType, quantity=None, is_singleton=False):
    """Render HTML for an asset with optional quantity."""
    try:
        name = eve_type.name
        icon_url = eve_type.icon_url()
        profile_url = eve_type.profile_url
    except AttributeError:
        name = icon_url = profile_url = ""
    return {
        "name": name,
        "icon_url": icon_url,
        "profile_url": profile_url,
        "quantity": quantity,
        "is_singleton": is_singleton,
    }


@register.inclusion_tag("structures/templatetags/list_tax_item.html")
def list_tax_item(title: str, value: float, has_access: bool):
    """Render HTML for list item."""
    return {"title": title, "value": value, "has_access": has_access}



import logging
from pathlib import Path

from esi.clients import EsiClientProvider

from app_utils.logging import LoggerAddTag

from . import __title__, __version__

logger = LoggerAddTag(logging.getLogger(__name__), __title__)
swagger_path = Path(__file__).parent / "swagger.json"
esi = EsiClientProvider(
    app_info_text=f"aa-structures v{__version__}", spec_file=swagger_path
)



import json

from django.test import TestCase
from django.utils.timezone import now

from app_utils.json import JSONDateTimeDecoder, JSONDateTimeEncoder


class TestJsonSerializer(TestCase):
    def test_encode_decode(self):
        my_dict = {"alpha": "hello", "bravo": now()}
        my_json = json.dumps(my_dict, cls=JSONDateTimeEncoder)
        my_dict_new = json.loads(my_json, cls=JSONDateTimeDecoder)
        self.assertDictEqual(my_dict, my_dict_new)



"""Eve Universe models"""

from enum import Enum

from django.db import models
from django.utils import translation
from django.utils.timezone import now
from django.utils.translation import gettext
from eveuniverse.core import dotlan, eveimageserver, eveitems

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from .. import __title__
from ..managers import EveEntityManager, EveSovereigntyMapManager, EveUniverseManager

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

NAMES_MAX_LENGTH = 255


class EveSovereigntyMap(models.Model):
    """Shows which alliance / corporation / faction owns a system

    Note: This model does not hold FKs to respective objects like
    EveSolarSystem to avoid having load all those object from ESI
    """

    solar_system_id = models.PositiveIntegerField(primary_key=True)
    alliance_id = models.PositiveIntegerField(
        blank=True,
        null=True,
        db_index=True,
        help_text="alliance who holds sov for this system",
    )
    corporation_id = models.PositiveIntegerField(
        blank=True,
        null=True,
        db_index=True,
        help_text="corporation who holds sov for this system",
    )
    faction_id = models.PositiveIntegerField(
        blank=True,
        null=True,
        db_index=True,
        help_text="faction who holds sov for this system",
    )
    last_updated = models.DateTimeField(
        default=None,
        null=True,
        blank=True,
        help_text="When this object was last updated from ESI",
        db_index=True,
    )

    objects = EveSovereigntyMapManager()

    def __str__(self):
        return str(self.solar_system_id)

    def __repr__(self):
        return "{}(solar_system_id='{}')".format(
            self.__class__.__name__, self.solar_system_id
        )


class EveSpaceType(str, Enum):
    UNKNOWN = "unknown"
    HIGHSEC = "highsec"
    LOWSEC = "lowsec"
    NULLSEC = "nullsec"
    W_SPACE = "w-space"

    @classmethod
    def from_solar_system(cls, eve_solar_system: "EveSolarSystem") -> "EveSpaceType":
        """returns the space type"""
        if eve_solar_system.is_null_sec:
            return cls.NULLSEC
        elif eve_solar_system.is_low_sec:
            return cls.LOWSEC
        elif eve_solar_system.is_high_sec:
            return cls.HIGHSEC
        elif eve_solar_system.is_w_space:
            return cls.W_SPACE
        return cls.UNKNOWN


class EsiNameLocalization(models.Model):
    """Base class for adding localization support for name field"""

    # language code mappings
    # Django, Eve Universe models, ESI
    LANG_CODES_MAPPING = (
        ("en", "en", "en-us"),
        ("de", "de", "de"),
        ("ko", "ko", "ko"),
        ("ru", "ru", "ru"),
        # ("zh-hans", "zh", "zh"),
    )
    LANG_CODES_DJANGO = 0
    LANG_CODES_MODEL = 1
    LANG_CODES_ESI = 2

    ESI_LANGUAGES = {x[2] for x in LANG_CODES_MAPPING}
    ESI_DEFAULT_LANGUAGE = "en-us"

    name_de = models.CharField(
        max_length=NAMES_MAX_LENGTH,
        blank=True,
        help_text="Eve Online name localized for German",
    )
    name_ko = models.CharField(
        max_length=NAMES_MAX_LENGTH,
        blank=True,
        help_text="Eve Online name localized for Korean",
    )
    name_ru = models.CharField(
        max_length=NAMES_MAX_LENGTH,
        blank=True,
        help_text="Eve Online name localized for Russian",
    )
    name_zh = models.CharField(
        max_length=NAMES_MAX_LENGTH,
        blank=True,
        help_text="Eve Online name localized for Chinese",
    )

    @property
    def name_localized(self):
        """returns the localized version of name for the current language

        will return the default if a translation does not exist
        """
        return self.name_localized_for_language(translation.get_language())

    def name_localized_for_language(self, language: str):
        """returns the localized version of name for the given language

        will return the default if a translation does not exist
        """
        lang_mapping = self._language_code_translation(language, self.LANG_CODES_DJANGO)
        if lang_mapping and (
            lang_mapping[self.LANG_CODES_ESI] != self.ESI_DEFAULT_LANGUAGE
        ):
            field_ext = lang_mapping[self.LANG_CODES_MODEL]
            field_name = ("name_%s" % field_ext) if field_ext else "name"
        else:
            field_name = "name"
        if not hasattr(self, field_name):
            raise NotImplementedError(
                'field name "%s" not found in %s' % (field_name, type(self).__name__)
            )
        name_translation = getattr(self, field_name)
        return name_translation if name_translation else self.name

    def _name_localized_generated(self, lang_code: str) -> str:
        raise NotImplementedError()

    class Meta:
        abstract = True

    @classmethod
    def _language_code_translation(cls, code: str, category_from: int) -> tuple:
        """translates language codes between systems"""
        result = None
        for mapping in cls.LANG_CODES_MAPPING:
            if mapping[category_from] == code:
                result = mapping
        return result


class EveUniverse(EsiNameLocalization, models.Model):
    """Base class for all EveUniverse models

    Eve Universe classes need to have a meta class defined: `EveUniverseMeta`

    Properties are:

    esi_pk: Name of the ESI property for the primary key, e.g. 'category_id'

    esi_method: name of the ESI method to be called for fetching objects

    children: dict of mapping between ESI dict and model class for
    children objects (e.g. planets for solar system), default is None

    fk_mappings: mapping of field names from model to ESI for FKs,
    default is None

    field_mappings: mapping of field names from model to ESI for non FKs,
    default is None

    has_esi_localization: True/False, whether this model gets translations from ESI,
    Default is True

    generate_localization: (optional) True/False, whether this model will
    generate localizations by itself, default is false

    """

    id = models.PositiveIntegerField(primary_key=True, help_text="Eve Online ID")
    name = models.CharField(max_length=NAMES_MAX_LENGTH, help_text="Eve Online name")
    last_updated = models.DateTimeField(
        default=None,
        null=True,
        blank=True,
        help_text="When this object was last updated from ESI",
        db_index=True,
    )

    objects = EveUniverseManager()

    class Meta:
        abstract = True

    def __repr__(self):
        return "{}(id={}, name='{}')".format(
            self.__class__.__name__, self.id, self.name
        )

    def __str__(self):
        return self.name

    def set_generated_translations(self):
        """updates localization fields with generated values if defined

        Purpose is to provide localized names for models where ESI does
        not provide localizations and where those names can be generated
        e.g. planets, moons

        Will look for _name_localized_generated() defined in the model
        and run it to set all localized names
        Does nothing if that method is not defined
        """
        if self._eve_universe_meta_attr("generate_localization"):
            for django_lc, field_ext, esi_lc in self.LANG_CODES_MAPPING:
                if esi_lc != self.ESI_DEFAULT_LANGUAGE:
                    field_name = "name_" + field_ext
                    setattr(self, field_name, self._name_localized_generated(django_lc))

    @classmethod
    def esi_pk(cls) -> str:
        """returns the name of the pk column on ESI that must exist"""
        return cls._eve_universe_meta_attr("esi_pk", is_mandatory=True)

    @classmethod
    def esi_method(cls) -> str:
        return cls._eve_universe_meta_attr("esi_method", is_mandatory=True)

    @classmethod
    def has_esi_localization(cls) -> bool:
        has_esi_localization = cls._eve_universe_meta_attr("has_esi_localization")
        return True if has_esi_localization is None else has_esi_localization

    @classmethod
    def child_mappings(cls) -> dict:
        """returns the mapping of children for this class"""
        mappings = cls._eve_universe_meta_attr("children")
        return mappings if mappings else dict()

    @classmethod
    def _field_names_not_pk(cls) -> set:
        """returns field names excl. PK, localization and auto created fields"""
        return {
            x.name
            for x in cls._meta.get_fields()
            if not x.auto_created
            and (not hasattr(x, "primary_key") or x.primary_key is False)
            and x.name not in {"language_code", "last_updated"}
            and "name_" not in x.name
        }

    @classmethod
    def _field_mappings(cls) -> dict:
        """returns the mappings for model fields vs. esi fields"""
        mappings = cls._eve_universe_meta_attr("field_mappings")
        return mappings if mappings else dict()

    @classmethod
    def _fk_mappings(cls) -> dict:
        """returns the foreign key mappings for this class

        'model field name': ('Foreign Key name on ESI', 'related model class')
        """

        def convert_to_esi_name(name: str, extra_fk_mappings: dict) -> str:
            if name in extra_fk_mappings:
                esi_name = extra_fk_mappings[name]
            else:
                esi_name = name.replace("eve_", "") + "_id"
            return esi_name

        extra_fk_mappings = cls._eve_universe_meta_attr("fk_mappings")
        if not extra_fk_mappings:
            extra_fk_mappings = {}

        mappings = {
            x.name: (convert_to_esi_name(x.name, extra_fk_mappings), x.related_model)
            for x in cls._meta.get_fields()
            if isinstance(x, models.ForeignKey)
        }
        return mappings

    @classmethod
    def map_esi_fields_to_model(cls, eve_data_objects: dict) -> dict:
        """maps ESi fields to model fields incl. translations if any
        returns the result as defaults dict
        """
        fk_mappings = cls._fk_mappings()
        field_mappings = cls._field_mappings()
        defaults = {"last_updated": now()}
        eve_data_obj = eve_data_objects[cls.ESI_DEFAULT_LANGUAGE]
        for key in cls._field_names_not_pk():
            if key in fk_mappings:
                esi_key, ParentClass = fk_mappings[key]
                value, _ = ParentClass.objects.get_or_create_esi(
                    id=eve_data_obj[esi_key]
                )
            else:
                if key in field_mappings:
                    mapping = field_mappings[key]
                    if len(mapping) != 2:
                        raise ValueError(
                            "Currently only supports mapping to 1-level " "nested dicts"
                        )
                    value = eve_data_obj[mapping[0]][mapping[1]]
                else:
                    value = eve_data_obj[key]

            defaults[key] = value

        # add translations if any
        if cls.has_esi_localization():
            for _, field_ext, esi_lc in cls.LANG_CODES_MAPPING:
                if esi_lc != cls.ESI_DEFAULT_LANGUAGE:
                    field_name = "name_" + field_ext
                    defaults[field_name] = eve_data_objects[esi_lc]["name"]

        return defaults

    @classmethod
    def _eve_universe_meta_attr(cls, attr_name: str, is_mandatory: bool = False):
        """returns value of an attribute from EveUniverseMeta or None"""
        if not hasattr(cls, "EveUniverseMeta"):
            raise ValueError("EveUniverseMeta not defined for class %s" % cls.__name__)

        if hasattr(cls.EveUniverseMeta, attr_name):
            value = getattr(cls.EveUniverseMeta, attr_name)
        else:
            value = None
            if is_mandatory:
                raise ValueError(
                    "Mandatory attribute EveUniverseMeta.%s not defined "
                    "for class %s" % (attr_name, cls.__name__)
                )
        return value


class EveCategory(EveUniverse):
    """category in Eve Online"""

    class EveUniverseMeta:
        esi_pk = "category_id"
        esi_method = "get_universe_categories_category_id"


class EveGroup(EveUniverse):
    """group in Eve Online"""

    eve_category = models.ForeignKey(
        EveCategory,
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        blank=True,
        related_name="eve_groups",
    )

    class EveUniverseMeta:
        esi_pk = "group_id"
        esi_method = "get_universe_groups_group_id"


class EveType(EveUniverse):
    """type in Eve Online"""

    eve_group = models.ForeignKey(
        EveGroup, on_delete=models.CASCADE, related_name="eve_types"
    )

    class EveUniverseMeta:
        esi_pk = "type_id"
        esi_method = "get_universe_types_type_id"

    @property
    def profile_url(self) -> str:
        return eveitems.type_url(self.id)

    def icon_url(self, size=64):
        return self.generic_icon_url(self.id, size)

    def render_url(self, size: int = 128) -> str:
        """return an image URL to this type as render"""
        return eveimageserver.type_render_url(self.id, size=size)

    @classmethod
    def generic_icon_url(cls, type_id: int, size: int = 64) -> str:
        return eveimageserver.type_icon_url(type_id=type_id, size=size)


class EveRegion(EveUniverse):
    """region in Eve Online"""

    class EveUniverseMeta:
        esi_pk = "region_id"
        esi_method = "get_universe_regions_region_id"

    @property
    def profile_url(self) -> str:
        """URL to display this type on the default third party webpage."""
        return dotlan.region_url(self.name)


class EveConstellation(EveUniverse):
    """constellation in Eve Online"""

    eve_region = models.ForeignKey(
        EveRegion, on_delete=models.CASCADE, related_name="eve_constellations"
    )

    class EveUniverseMeta:
        esi_pk = "constellation_id"
        esi_method = "get_universe_constellations_constellation_id"


class EveSolarSystem(EveUniverse):
    """solar system in Eve Online"""

    eve_constellation = models.ForeignKey(
        EveConstellation, on_delete=models.CASCADE, related_name="eve_solar_systems"
    )
    security_status = models.FloatField()

    class EveUniverseMeta:
        esi_pk = "system_id"
        esi_method = "get_universe_systems_system_id"
        children = {"planets": "EvePlanet"}

    @property
    def is_high_sec(self) -> bool:
        """returns True if this solar system is in high sec, else False"""
        return round(self.security_status, 1) >= 0.5

    @property
    def is_low_sec(self) -> bool:
        """returns True if this solar system is in low sec, else False"""
        return 0 < round(self.security_status, 1) < 0.5

    @property
    def is_null_sec(self) -> bool:
        """returns True if this solar system is in null sec, else False"""
        return round(self.security_status, 1) <= 0 and not self.is_w_space

    @property
    def is_w_space(self) -> bool:
        """returns True if this solar system is in wormhole space, else False"""
        return 31000000 <= self.id < 32000000

    @property
    def profile_url(self) -> str:
        """URL to display this type on the default third party webpage."""
        return dotlan.solar_system_url(self.name)


class EvePlanet(EveUniverse):
    """planet in Eve Online"""

    position_x = models.FloatField(
        null=True, default=None, blank=True, help_text="x position in the solar system"
    )
    position_y = models.FloatField(
        null=True, default=None, blank=True, help_text="y position in the solar system"
    )
    position_z = models.FloatField(
        null=True, default=None, blank=True, help_text="z position in the solar system"
    )
    eve_solar_system = models.ForeignKey(
        EveSolarSystem, on_delete=models.CASCADE, related_name="eve_planets"
    )
    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE)

    def _name_localized_generated(self, language: str) -> str:
        """returns a generated localized planet name for the given language"""
        name_localized = self.name.replace(
            self.eve_solar_system.name,
            self.eve_solar_system.name_localized_for_language(language),
        )
        return name_localized

    class EveUniverseMeta:
        esi_pk = "planet_id"
        esi_method = "get_universe_planets_planet_id"
        fk_mappings = {"eve_solar_system": "system_id"}
        field_mappings = {
            "position_x": ("position", "x"),
            "position_y": ("position", "y"),
            "position_z": ("position", "z"),
        }
        has_esi_localization = False
        generate_localization = True


class EveMoon(EveUniverse):
    """ "moon in Eve Online"""

    position_x = models.FloatField(
        null=True, default=None, blank=True, help_text="x position in the solar system"
    )
    position_y = models.FloatField(
        null=True, default=None, blank=True, help_text="y position in the solar system"
    )
    position_z = models.FloatField(
        null=True, default=None, blank=True, help_text="z position in the solar system"
    )
    # TODO: Variation from eveuniverse
    eve_solar_system = models.ForeignKey(
        EveSolarSystem, on_delete=models.CASCADE, related_name="eve_moons"
    )

    def _name_localized_generated(self, language: str):
        """returns a generated localized moon name for the given language"""
        with translation.override(language):
            name_localized = self.name.replace(
                self.eve_solar_system.name, self.eve_solar_system.name_localized
            ).replace("Moon", gettext("Moon"))
        return name_localized

    class EveUniverseMeta:
        esi_pk = "moon_id"
        esi_method = "get_universe_moons_moon_id"
        fk_mappings = {"eve_solar_system": "system_id"}
        field_mappings = {
            "position_x": ("position", "x"),
            "position_y": ("position", "y"),
            "position_z": ("position", "z"),
        }
        has_esi_localization = False
        generate_localization = True


class EveEntity(models.Model):
    """An EVE entity like a character or an alliance."""

    class Category(models.IntegerChoices):
        CHARACTER = 1, "character"
        CORPORATION = 2, "corporation"
        ALLIANCE = 3, "alliance"
        FACTION = 4, "faction"
        OTHER = 5, "other"

        @classmethod
        def from_esi_name(cls, esi_name: str) -> "EveEntity.Category":
            """Returns category for given ESI name."""
            for choice in cls.choices:
                if esi_name == choice[1]:
                    return cls(choice[0])
            return cls.OTHER

    id = models.PositiveIntegerField(primary_key=True, help_text="Eve Online ID")
    category = models.IntegerField(choices=Category.choices)
    name = models.CharField(max_length=255, null=True, default=None, blank=True)

    objects = EveEntityManager()

    def __str__(self) -> str:
        return str(self.name)

    def __repr__(self) -> str:
        return "{}(id={}, category='{}', name='{}')".format(
            self.__class__.__name__, self.id, self.get_category_display(), self.name
        )

    @property
    def profile_url(self) -> str:
        """Returns link to website with profile info about this entity."""
        if self.category == self.Category.CORPORATION:
            url = dotlan.corporation_url(self.name)
        elif self.category == self.Category.ALLIANCE:
            url = dotlan.alliance_url(self.name)
        else:
            url = ""
        return url

    def icon_url(self, size: int = 32) -> str:
        if self.category == self.Category.ALLIANCE:
            return eveimageserver.alliance_logo_url(self.id, size)
        elif (
            self.category == self.Category.CORPORATION
            or self.category == self.Category.FACTION
        ):
            return eveimageserver.corporation_logo_url(self.id, size)
        elif self.category == self.Category.CHARACTER:
            return eveimageserver.character_portrait_url(self.id, size)
        else:
            raise NotImplementedError()



# Generated by Django 3.2.11 on 2022-01-28 18:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("structures", "0033_fix_notification_type_fuel_alert"),
    ]

    operations = [
        migrations.AddField(
            model_name="notification",
            name="structures",
            field=models.ManyToManyField(
                help_text="Structures this notification is about (if any)",
                related_name="notifications",
                to="structures.Structure",
            ),
        ),
        migrations.AddField(
            model_name="structure",
            name="webhooks",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Webhooks for sending notifications to. If any webhook is enabled, these will be used instead of the webhooks defined for the respective owner. If no webhook is enabled the owner's setting will be used. ",
                related_name="structures",
                to="structures.Webhook",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="ping_groups",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Groups to be pinged for each notification. ",
                related_name="_structures_owner_ping_groups_+",
                to="auth.Group",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="webhooks",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Notifications are sent to these webhooks. ",
                related_name="owners",
                to="structures.Webhook",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="tags",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="List of tags for this structure. ",
                related_name="structures",
                to="structures.StructureTag",
            ),
        ),
    ]



from django.core.management.base import BaseCommand
from eveuniverse.models import EveCategory as EveCategory2
from eveuniverse.models import EveConstellation as EveConstellation2
from eveuniverse.models import EveEntity as EveEntity2
from eveuniverse.models import EveGroup as EveGroup2
from eveuniverse.models import EveMoon as EveMoon2
from eveuniverse.models import EvePlanet as EvePlanet2
from eveuniverse.models import EveRegion as EveRegion2
from eveuniverse.models import EveSolarSystem as EveSolarSystem2
from eveuniverse.models import EveType as EveType2

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from ... import __title__
from ...models import (
    EveCategory,
    EveConstellation,
    EveEntity,
    EveGroup,
    EveMoon,
    EvePlanet,
    EveRegion,
    EveSolarSystem,
    EveType,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def get_input(text):
    """wrapped input to enable unit testing / patching"""
    return input(text)


class Command(BaseCommand):
    help = "Preload missing eveuniverse objects in preparation for migrating to eveuniverse with release 2."

    def handle(self, *args, **options):
        models_map = {
            EveEntity: EveEntity2,
            EveCategory: EveCategory2,
            EveGroup: EveGroup2,
            EveType: EveType2,
            EveRegion: EveRegion2,
            EveConstellation: EveConstellation2,
            EveSolarSystem: EveSolarSystem2,
            EvePlanet: EvePlanet2,
            EveMoon: EveMoon2,
        }
        for OldModel, NewModel in models_map.items():
            ids_target = set(OldModel.objects.values_list("id", flat=True))
            ids_current = set(NewModel.objects.values_list("id", flat=True))
            ids_diff = ids_target.difference(ids_current)
            if ids_diff:
                self.stdout.write(
                    f"{OldModel.__name__}: Need to fetch {len(ids_diff)} "
                    "missing object(s) from ESI."
                )
            else:
                self.stdout.write(f"{OldModel.__name__}: No outstanding objects.")
            NewModel.objects.bulk_get_or_create_esi(ids=ids_target)
        self.stdout.write(self.style.SUCCESS("DONE"))



"""Notification related models"""

import datetime as dt
import math
from typing import Optional, Tuple

import dhooks_lite
import yaml
from multiselectfield import MultiSelectField
from requests.exceptions import HTTPError

from django.contrib.auth.models import Group
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.utils import translation
from django.utils.functional import classproperty
from django.utils.timezone import now
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveEntity as EveEntity2

from allianceauth.services.hooks import get_extension_logger
from app_utils.datetime import (
    DATETIME_FORMAT,
    ldap_time_2_datetime,
    ldap_timedelta_2_timedelta,
)
from app_utils.django import app_labels
from app_utils.logging import LoggerAddTag
from app_utils.urls import static_file_absolute_url

from .. import __title__
from ..app_settings import (  # STRUCTURES_NOTIFICATION_DISABLE_ESI_FUEL_ALERTS,
    STRUCTURES_ADD_TIMERS,
    STRUCTURES_DEFAULT_LANGUAGE,
    STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS,
    STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED,
    STRUCTURES_NOTIFICATION_SET_AVATAR,
    STRUCTURES_REPORT_NPC_ATTACKS,
    STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
)
from ..constants import EveCategoryId, EveCorporationId, EveTypeId
from ..core import starbases
from ..managers import GeneratedNotificationManager, NotificationManager, WebhookManager
from ..webhooks.models import WebhookBase
from .eveuniverse import EveEntity, EveMoon, EvePlanet, EveSolarSystem
from .structures import Structure

if "timerboard" in app_labels():
    from allianceauth.timerboard.models import Timer as AuthTimer

    has_auth_timers = True
else:
    has_auth_timers = False

if "structuretimers" in app_labels():
    from structuretimers.models import Timer

    from eveuniverse.models import EveSolarSystem as EveSolarSystem2
    from eveuniverse.models import EveType as EveType2

    has_structure_timers = True
else:
    has_structure_timers = False

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

# Supported languages
LANGUAGES = (
    ("en", _("English")),
    ("de", _("German")),
    ("es", _("Spanish")),
    ("zh-hans", _("Chinese Simplified")),
    ("ru", _("Russian")),
    ("ko", _("Korean")),
)


class NotificationType(models.TextChoices):
    """Definition of all supported notification types."""

    # upwell structures
    STRUCTURE_ANCHORING = "StructureAnchoring", _("Upwell structure anchoring")
    STRUCTURE_ONLINE = "StructureOnline", _("Upwell structure went online")
    STRUCTURE_SERVICES_OFFLINE = "StructureServicesOffline", _(
        "Upwell structure services went offline"
    )
    STRUCTURE_WENT_HIGH_POWER = "StructureWentHighPower", _(
        "Upwell structure went high power"
    )
    STRUCTURE_WENT_LOW_POWER = "StructureWentLowPower", _(
        "Upwell structure went low power"
    )
    STRUCTURE_UNANCHORING = "StructureUnanchoring", _("Upwell structure unanchoring")
    STRUCTURE_FUEL_ALERT = "StructureFuelAlert", _("Upwell structure fuel alert")
    STRUCTURE_REFUELED_EXTRA = "StructureRefueledExtra", _("Upwell structure refueled")
    STRUCTURE_JUMP_FUEL_ALERT = "StructureJumpFuelAlert", _(
        "Upwell structure jump fuel alert"
    )
    STRUCTURE_UNDER_ATTACK = "StructureUnderAttack", _(
        "Upwell structure is under attack"
    )
    STRUCTURE_LOST_SHIELD = "StructureLostShields", _("Upwell structure lost shields")
    STRUCTURE_LOST_ARMOR = "StructureLostArmor", _("Upwell structure lost armor")
    STRUCTURE_DESTROYED = "StructureDestroyed", _("Upwell structure destroyed")

    STRUCTURE_REINFORCE_CHANGED = "StructuresReinforcementChanged", _(
        "Upwell structure reinforcement time changed"
    )
    OWNERSHIP_TRANSFERRED = "OwnershipTransferred", _(
        "Upwell structure ownership transferred"
    )

    # customs offices
    ORBITAL_ATTACKED = "OrbitalAttacked", _("Customs office attacked")
    ORBITAL_REINFORCED = "OrbitalReinforced", _("Customs office reinforced")

    # starbases
    TOWER_ALERT_MSG = "TowerAlertMsg", _("Starbase attacked")
    TOWER_RESOURCE_ALERT_MSG = "TowerResourceAlertMsg", _("Starbase fuel alert")
    TOWER_REFUELED_EXTRA = "TowerRefueledExtra", _("Starbase refueled (BETA)")
    TOWER_REINFORCED_EXTRA = "TowerReinforcedExtra", _("Starbase reinforced (BETA)")

    # moon mining
    MOONMINING_EXTRACTION_STARTED = "MoonminingExtractionStarted", _(
        "Moonmining extraction started"
    )
    MOONMINING_LASER_FIRED = "MoonminingLaserFired", _("Moonmining laser fired")
    MOONMINING_EXTRACTION_CANCELLED = "MoonminingExtractionCancelled", _(
        "Moonmining extraction cancelled"
    )
    MOONMINING_EXTRACTION_FINISHED = "MoonminingExtractionFinished", _(
        "Moonmining extraction finished"
    )
    MOONMINING_AUTOMATIC_FRACTURE = "MoonminingAutomaticFracture", _(
        "Moonmining automatic fracture triggered"
    )

    # sov
    SOV_STRUCTURE_REINFORCED = "SovStructureReinforced", _(
        "Sovereignty structure reinforced"
    )
    SOV_STRUCTURE_DESTROYED = "SovStructureDestroyed", _(
        "Sovereignty structure destroyed"
    )
    SOV_ENTOSIS_CAPTURE_STARTED = "EntosisCaptureStarted", _(
        "Sovereignty entosis capture started"
    )
    SOV_COMMAND_NODE_EVENT_STARTED = "SovCommandNodeEventStarted", _(
        "Sovereignty command node event started"
    )
    SOV_ALL_CLAIM_ACQUIRED_MSG = "SovAllClaimAquiredMsg", _(
        "Sovereignty claim acknowledgment"
    )
    SOV_ALL_CLAIM_LOST_MSG = "SovAllClaimLostMsg", _("Sovereignty lost")
    SOV_ALL_ANCHORING_MSG = "AllAnchoringMsg", _(
        "Structure anchoring in alliance space"
    )

    # wars
    WAR_WAR_DECLARED = "WarDeclared", _("War declared")
    WAR_ALLY_JOINED_WAR_AGGRESSOR_MSG = "AllyJoinedWarAggressorMsg", _(
        "War ally joined aggressor"
    )
    WAR_ALLY_JOINED_WAR_AllY_MSG = "AllyJoinedWarAllyMsg", _("War ally joined ally")
    WAR_ALLY_JOINED_WAR_DEFENDER_MSG = "AllyJoinedWarDefenderMsg", _(
        "War ally joined defender"
    )
    WAR_WAR_ADOPTED = "WarAdopted", _("War adopted")
    WAR_WAR_INHERITED = "WarInherited", _("War inherited")
    WAR_CORP_WAR_SURRENDER_MSG = "CorpWarSurrenderMsg", _("War party surrendered")
    WAR_WAR_RETRACTED_BY_CONCORD = "WarRetractedByConcord", _(
        "War retracted by Concord"
    )
    WAR_CORPORATION_BECAME_ELIGIBLE = "CorpBecameWarEligible", _(
        "War corporation became eligible"
    )
    WAR_CORPORATION_NO_LONGER_ELIGIBLE = "CorpNoLongerWarEligible", _(
        "War corporation no longer eligible"
    )
    WAR_WAR_SURRENDER_OFFER_MSG = "WarSurrenderOfferMsg", _("War surrender offered")

    # corporation membership
    CORP_APP_NEW_MSG = "CorpAppNewMsg", _("Character submitted application")
    CORP_APP_INVITED_MSG = "CorpAppInvitedMsg", _(
        "Character invited to join corporation"
    )
    CORP_APP_REJECT_CUSTOM_MSG = "CorpAppRejectCustomMsg", _(
        "Corp application rejected"
    )
    CHAR_APP_WITHDRAW_MSG = "CharAppWithdrawMsg", _("Character withdrew application")
    CHAR_APP_ACCEPT_MSG = "CharAppAcceptMsg", _("Character joins corporation")
    CHAR_LEFT_CORP_MSG = "CharLeftCorpMsg", _("Character leaves corporation")

    # billing
    BILLING_BILL_OUT_OF_MONEY_MSG = "BillOutOfMoneyMsg", _("Bill out of money")
    BILLING_I_HUB_BILL_ABOUT_TO_EXPIRE = (
        "InfrastructureHubBillAboutToExpire",
        _("I-HUB bill about to expire"),
    )
    BILLING_I_HUB_DESTROYED_BY_BILL_FAILURE = (
        "IHubDestroyedByBillFailure",
        _("I_HUB destroyed by bill failure"),
    )

    @classproperty
    def esi_notifications(cls) -> set:
        return set(cls.values) - cls.generated_notifications

    @classproperty
    def generated_notifications(cls) -> set:
        return {
            cls.STRUCTURE_JUMP_FUEL_ALERT,
            cls.STRUCTURE_REFUELED_EXTRA,
            cls.TOWER_REFUELED_EXTRA,
            cls.TOWER_REINFORCED_EXTRA,
        }

    @classproperty
    def webhook_defaults(cls) -> list:
        """List of default notifications for new webhooks."""
        return [
            cls.STRUCTURE_ANCHORING,
            cls.STRUCTURE_DESTROYED,
            cls.STRUCTURE_FUEL_ALERT,
            cls.STRUCTURE_LOST_ARMOR,
            cls.STRUCTURE_LOST_SHIELD,
            cls.STRUCTURE_ONLINE,
            cls.STRUCTURE_SERVICES_OFFLINE,
            cls.STRUCTURE_UNDER_ATTACK,
            cls.STRUCTURE_WENT_HIGH_POWER,
            cls.STRUCTURE_WENT_LOW_POWER,
            cls.ORBITAL_ATTACKED,
            cls.ORBITAL_REINFORCED,
            cls.TOWER_ALERT_MSG,
            cls.TOWER_RESOURCE_ALERT_MSG,
            cls.SOV_STRUCTURE_REINFORCED,
            cls.SOV_STRUCTURE_DESTROYED,
        ]

    @classproperty
    def relevant_for_timerboard(cls) -> set:
        """Notification types that can create timers."""
        return {
            cls.STRUCTURE_LOST_SHIELD,
            cls.STRUCTURE_LOST_ARMOR,
            cls.ORBITAL_REINFORCED,
            cls.MOONMINING_EXTRACTION_STARTED,
            cls.MOONMINING_EXTRACTION_CANCELLED,
            cls.SOV_STRUCTURE_REINFORCED,
            cls.TOWER_REINFORCED_EXTRA,
        }

    @classproperty
    def relevant_for_alliance_level(cls) -> set:
        """Notification types that require the alliance level flag."""
        return {
            # billing
            cls.BILLING_BILL_OUT_OF_MONEY_MSG,
            cls.BILLING_I_HUB_DESTROYED_BY_BILL_FAILURE,
            cls.BILLING_I_HUB_BILL_ABOUT_TO_EXPIRE,
            # sov
            cls.SOV_ENTOSIS_CAPTURE_STARTED,
            cls.SOV_COMMAND_NODE_EVENT_STARTED,
            cls.SOV_ALL_CLAIM_ACQUIRED_MSG,
            cls.SOV_STRUCTURE_REINFORCED,
            cls.SOV_STRUCTURE_DESTROYED,
            cls.SOV_ALL_CLAIM_LOST_MSG,
            cls.SOV_ALL_ANCHORING_MSG,
            # wars
            cls.WAR_ALLY_JOINED_WAR_AGGRESSOR_MSG,
            cls.WAR_ALLY_JOINED_WAR_AllY_MSG,
            cls.WAR_ALLY_JOINED_WAR_DEFENDER_MSG,
            cls.WAR_CORP_WAR_SURRENDER_MSG,
            cls.WAR_CORPORATION_BECAME_ELIGIBLE,
            cls.WAR_CORPORATION_NO_LONGER_ELIGIBLE,
            cls.WAR_WAR_ADOPTED,
            cls.WAR_WAR_DECLARED,
            cls.WAR_WAR_INHERITED,
            cls.WAR_WAR_RETRACTED_BY_CONCORD,
            cls.WAR_WAR_SURRENDER_OFFER_MSG,
        }

    @classproperty
    def relevant_for_moonmining(cls) -> set:
        """Notification types about moon mining."""
        return {
            cls.MOONMINING_EXTRACTION_STARTED,
            cls.MOONMINING_EXTRACTION_CANCELLED,
            cls.MOONMINING_LASER_FIRED,
            cls.MOONMINING_EXTRACTION_FINISHED,
            cls.MOONMINING_AUTOMATIC_FRACTURE,
        }

    @classproperty
    def structure_related(cls) -> set:
        """Notification types that are related to a structure."""
        return {
            cls.STRUCTURE_ONLINE,
            cls.STRUCTURE_FUEL_ALERT,
            cls.STRUCTURE_JUMP_FUEL_ALERT,
            cls.STRUCTURE_REFUELED_EXTRA,
            cls.STRUCTURE_SERVICES_OFFLINE,
            cls.STRUCTURE_WENT_LOW_POWER,
            cls.STRUCTURE_WENT_HIGH_POWER,
            cls.STRUCTURE_UNANCHORING,
            cls.STRUCTURE_UNDER_ATTACK,
            cls.STRUCTURE_LOST_SHIELD,
            cls.STRUCTURE_LOST_ARMOR,
            cls.STRUCTURE_DESTROYED,
            cls.OWNERSHIP_TRANSFERRED,
            cls.STRUCTURE_ANCHORING,
            cls.MOONMINING_EXTRACTION_STARTED,
            cls.MOONMINING_EXTRACTION_FINISHED,
            cls.MOONMINING_AUTOMATIC_FRACTURE,
            cls.MOONMINING_EXTRACTION_CANCELLED,
            cls.MOONMINING_LASER_FIRED,
            cls.STRUCTURE_REINFORCE_CHANGED,
            cls.ORBITAL_ATTACKED,
            cls.ORBITAL_REINFORCED,
            cls.TOWER_ALERT_MSG,
            cls.TOWER_RESOURCE_ALERT_MSG,
            cls.TOWER_REFUELED_EXTRA,
            cls.TOWER_REINFORCED_EXTRA,
        }

    @classproperty
    def relevant_for_forwarding(cls) -> set:
        """Notification types that are forwarded to Discord."""
        my_set = set(cls.values_enabled)
        # if STRUCTURES_NOTIFICATION_DISABLE_ESI_FUEL_ALERTS:
        #     my_set.discard(cls.STRUCTURE_FUEL_ALERT)
        #     my_set.discard(cls.TOWER_RESOURCE_ALERT_MSG)
        return my_set

    @classproperty
    def values_enabled(cls) -> set:
        """Values of enabled notif types only."""
        my_set = set(cls.values)
        if not STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS:
            my_set.discard(cls.STRUCTURE_REFUELED_EXTRA)
            my_set.discard(cls.TOWER_REFUELED_EXTRA)
        return my_set

    @classproperty
    def choices_enabled(cls) -> list:
        """Choices list containing enabled notif types only."""
        return [choice for choice in cls.choices if choice[0] in cls.values_enabled]


def get_default_notification_types():
    """DEPRECATED: generates a set of all existing notification types as default.
    Required to support older migrations."""
    return tuple(sorted([str(x[0]) for x in NotificationType.values]))


class Webhook(WebhookBase):
    """A destination for forwarding notification alerts."""

    notification_types = MultiSelectField(
        choices=NotificationType.choices,
        default=NotificationType.webhook_defaults,
        help_text=(
            "select which type of notifications should be forwarded to this webhook"
        ),
    )
    language_code = models.CharField(
        max_length=8,
        choices=LANGUAGES,
        default=None,
        null=True,
        blank=True,
        verbose_name="language",
        help_text="language of notifications send to this webhook",
    )
    is_default = models.BooleanField(
        default=False,
        help_text=(
            "whether owners have this webhook automatically pre-set when created"
        ),
    )
    has_default_pings_enabled = models.BooleanField(
        default=True,
        help_text=(
            "to enable or disable pinging of notifications for this webhook "
            "e.g. with @everyone and @here"
        ),
    )
    ping_groups = models.ManyToManyField(
        Group,
        default=None,
        blank=True,
        related_name="+",
        help_text="Groups to be pinged for each notification - ",
    )
    objects = WebhookManager()

    @staticmethod
    def text_bold(text) -> str:
        """Format the given text in bold."""
        return f"**{text}**" if text else ""


class NotificationBase(models.Model):
    """Base model for a notification."""

    # event type structure map
    MAP_CAMPAIGN_EVENT_2_TYPE_ID = {
        1: EveTypeId.TCU,
        2: EveTypeId.IHUB,
    }
    MAP_TYPE_ID_2_TIMER_STRUCTURE_NAME = {
        EveTypeId.CUSTOMS_OFFICE: "POCO",
        EveTypeId.TCU: "TCU",
        EveTypeId.IHUB: "I-HUB",
    }

    is_sent = models.BooleanField(
        default=False,
        help_text="True when this notification has been forwarded to Discord",
    )
    is_timer_added = models.BooleanField(
        null=True,
        default=False,
        help_text="True when a timer has been added for this notification",
    )
    notif_type = models.CharField(
        max_length=100,
        default="",
        db_index=True,
        verbose_name="type",
        help_text="type of this notification",
    )
    owner = models.ForeignKey(
        "Owner",
        on_delete=models.CASCADE,
        help_text="Corporation that owns this notification",
    )
    structures = models.ManyToManyField(
        Structure, help_text="Structures this notification is about (if any)"
    )

    class Meta:
        abstract = True

    def __str__(self) -> str:
        return f"{self.notification_id}:{self.notif_type}"

    def __repr__(self) -> str:
        return "%s(notification_id=%d, owner='%s', notif_type='%s')" % (
            self.__class__.__name__,
            self.notification_id,
            self.owner,
            self.notif_type,
        )

    @property
    def is_alliance_level(self) -> bool:
        """Whether this is an alliance level notification."""
        return self.notif_type in NotificationType.relevant_for_alliance_level

    @property
    def can_be_rendered(self) -> bool:
        """Whether this notification can be rendered in Discord."""
        return self.notif_type in NotificationType.values

    @property
    def can_have_timer(self) -> bool:
        """Whether this notification can have a timer."""
        return self.notif_type in NotificationType.relevant_for_timerboard

    @property
    def is_structure_related(self) -> bool:
        """Weather this notification related to a structure."""
        return self.notif_type in NotificationType.structure_related

    def send_to_configured_webhooks(
        self,
        ping_type_override: Webhook.PingType = None,
        use_color_override: bool = False,
        color_override: int = None,
    ) -> Optional[bool]:
        """Send this notification to all active webhooks which have this
        notification type configured
        and apply filter for NPC attacks and alliance level if needed.

        Returns True, if notifications has been successfully send to webhooks
        Returns None, if owner has no fitting webhook
        Returns False, if sending to any webhooks failed
        """
        if self.filter_for_npc_attacks():
            logger.debug("%s: Will not send NPC attacks", self)
            return None
        if self.filter_for_alliance_level():
            logger.debug(
                "%s: Alliance level notifications are not enabled for this owner", self
            )
            return None
        webhooks_qs = self.relevant_webhooks()
        if not webhooks_qs.exists():
            logger.debug("%s: No relevant webhook found", self)
            return None
        if ping_type_override:
            self.ping_type_override = ping_type_override
        if use_color_override:
            self.color_override = color_override
        success = True
        for webhook in webhooks_qs:
            success &= self.send_to_webhook(webhook)
        return success

    def filter_for_npc_attacks(self) -> bool:
        """True when notification to be filtered out due to npc attacks."""
        return not STRUCTURES_REPORT_NPC_ATTACKS and self.is_npc_attacking()

    def filter_for_alliance_level(self) -> bool:
        """True when notification to be filtered out due to alliance level."""
        return (
            self.is_alliance_level
            and self.owner.corporation.alliance is not None
            and not self.owner.is_alliance_main
        )

    def relevant_webhooks(self) -> models.QuerySet:
        """Determine relevant webhooks matching this notification type."""
        if not self.is_structure_related:
            structures_qs = None
        else:
            structures_qs = self.calc_related_structures()
        if structures_qs and structures_qs.filter(webhooks__isnull=False).count() == 1:
            webhooks_qs = structures_qs.first().webhooks.filter(
                notification_types__contains=self.notif_type, is_active=True
            )
        else:
            webhooks_qs = self.owner.webhooks.filter(
                notification_types__contains=self.notif_type, is_active=True
            )
        return webhooks_qs

    def calc_related_structures(self) -> models.QuerySet[Structure]:
        """Identify structures this notification is related to.

        Returns:
        - structures if any found or empty list if there are no related structures
        """
        parsed_text = self.parsed_text()
        if not parsed_text:
            return Structure.objects.none()
        if self.notif_type in {
            NotificationType.STRUCTURE_ONLINE,
            NotificationType.STRUCTURE_FUEL_ALERT,
            NotificationType.STRUCTURE_JUMP_FUEL_ALERT,
            NotificationType.STRUCTURE_REFUELED_EXTRA,
            NotificationType.STRUCTURE_SERVICES_OFFLINE,
            NotificationType.STRUCTURE_WENT_LOW_POWER,
            NotificationType.STRUCTURE_WENT_HIGH_POWER,
            NotificationType.STRUCTURE_UNANCHORING,
            NotificationType.STRUCTURE_UNDER_ATTACK,
            NotificationType.STRUCTURE_LOST_SHIELD,
            NotificationType.STRUCTURE_LOST_ARMOR,
            NotificationType.STRUCTURE_DESTROYED,
            NotificationType.OWNERSHIP_TRANSFERRED,
            NotificationType.STRUCTURE_ANCHORING,
            NotificationType.MOONMINING_EXTRACTION_STARTED,
            NotificationType.MOONMINING_EXTRACTION_FINISHED,
            NotificationType.MOONMINING_AUTOMATIC_FRACTURE,
            NotificationType.MOONMINING_EXTRACTION_CANCELLED,
            NotificationType.MOONMINING_LASER_FIRED,
        }:
            structure_id = parsed_text.get("structureID")
            return Structure.objects.filter(id=structure_id)
        elif self.notif_type == NotificationType.STRUCTURE_REINFORCE_CHANGED:
            structure_ids = [
                structure_info[0] for structure_info in parsed_text["allStructureInfo"]
            ]
            return Structure.objects.filter(id__in=structure_ids)

        elif self.notif_type in {
            NotificationType.ORBITAL_ATTACKED,
            NotificationType.ORBITAL_REINFORCED,
        }:
            return Structure.objects.filter(
                eve_planet_id=parsed_text["planetID"], eve_type_id=parsed_text["typeID"]
            )

        elif self.notif_type in {
            NotificationType.TOWER_ALERT_MSG,
            NotificationType.TOWER_RESOURCE_ALERT_MSG,
            NotificationType.TOWER_REFUELED_EXTRA,
        }:
            return Structure.objects.filter(
                eve_moon_id=parsed_text["moonID"], eve_type_id=parsed_text["typeID"]
            )
        return Structure.objects.none()

    def send_to_webhook(self, webhook: Webhook) -> bool:
        """Sends this notification to the configured webhook.

        returns True if successful, else False
        """
        logger.info("%s: Trying to sent to webhook: %s", self, webhook)
        try:
            embed, ping_type = self._generate_embed(webhook.language_code)
        except (OSError, NotImplementedError) as ex:
            logger.warning("%s: Failed to generate embed: %s", self, ex, exc_info=True)
            return False
        if webhook.has_default_pings_enabled and self.owner.has_default_pings_enabled:
            if ping_type == Webhook.PingType.EVERYONE:
                content = "@everyone"
            elif ping_type == Webhook.PingType.HERE:
                content = "@here"
            else:
                content = ""
        else:
            content = ""
        if webhook.ping_groups.count() > 0 or self.owner.ping_groups.count() > 0:
            if "discord" in app_labels():
                DiscordUser = self._import_discord()

                groups = set(self.owner.ping_groups.all()) | set(
                    webhook.ping_groups.all()
                )
                for group in groups:
                    try:
                        role = DiscordUser.objects.group_to_role(group)
                    except HTTPError:
                        logger.warning("Failed to get Discord roles", exc_info=True)
                    else:
                        if role:
                            content += f" <@&{role['id']}>"

        username, avatar_url = self._gen_avatar()
        new_queue_size = webhook.send_message(
            content=content, embeds=[embed], username=username, avatar_url=avatar_url
        )
        success = new_queue_size > 0
        if success and not self.is_temporary:
            self.is_sent = True
            self.save()
        return success

    def _generate_embed(
        self, language_code: str
    ) -> Tuple[dhooks_lite.Embed, Webhook.PingType]:
        """Generates a Discord embed for this notification."""
        from ..core.notification_embeds import NotificationBaseEmbed

        logger.info("Creating embed with language = %s" % language_code)
        with translation.override(language_code):
            notification_embed = NotificationBaseEmbed.create(self)
            embed = notification_embed.generate_embed()
            return embed, notification_embed.ping_type

    @staticmethod
    def _import_discord() -> object:
        from allianceauth.services.modules.discord.models import DiscordUser

        return DiscordUser

    def _gen_avatar(self) -> Tuple[str, str]:
        if STRUCTURES_NOTIFICATION_SET_AVATAR:
            username = "Notifications"
            avatar_url = static_file_absolute_url("structures/img/structures_logo.png")
        else:
            username = None
            avatar_url = None

        return username, avatar_url

    def add_or_remove_timer(self) -> bool:
        """Add/remove a timer related to this notification for some types.

        Returns True when timers where added or removed, else False
        """
        if not STRUCTURES_ADD_TIMERS:
            return False

        if (
            not has_auth_timers and not has_structure_timers
        ) or self.notif_type not in NotificationType.relevant_for_timerboard:
            return False

        parsed_text = self.parsed_text()
        try:
            with translation.override(STRUCTURES_DEFAULT_LANGUAGE):
                if self.notif_type in [
                    NotificationType.STRUCTURE_LOST_ARMOR,
                    NotificationType.STRUCTURE_LOST_SHIELD,
                ]:
                    timer_processed = self._gen_timer_structure_reinforcement(
                        parsed_text
                    )
                elif self.notif_type == NotificationType.SOV_STRUCTURE_REINFORCED:
                    timer_processed = self._gen_timer_sov_reinforcements(parsed_text)
                elif self.notif_type == NotificationType.ORBITAL_REINFORCED:
                    timer_processed = self._gen_timer_orbital_reinforcements(
                        parsed_text
                    )
                elif self.notif_type in [
                    NotificationType.MOONMINING_EXTRACTION_STARTED,
                    NotificationType.MOONMINING_EXTRACTION_CANCELLED,
                ]:
                    if not STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED:
                        timer_processed = None
                    else:
                        timer_processed = self._gen_timer_moon_extraction(parsed_text)
                elif self.notif_type == NotificationType.TOWER_REINFORCED_EXTRA:
                    timer_processed = self._gen_timer_tower_reinforcements(parsed_text)
                else:
                    raise NotImplementedError()

            if timer_processed:
                logger.info("%s: Created timer for notification", self.notification_id)

            self.is_timer_added = timer_processed
            self.save()

        except OSError as ex:
            logger.warning(
                "%s: Failed to add timer from notification: %s",
                self,
                ex,
                exc_info=True,
            )

        return timer_processed

    def _gen_timer_structure_reinforcement(self, parsed_text: str) -> bool:
        """Generate timer for structure reinforcements"""
        token = self.owner.fetch_token()
        structure_obj, _ = Structure.objects.get_or_create_esi(
            id=parsed_text["structureID"], token=token
        )
        eve_time = self.timestamp + ldap_timedelta_2_timedelta(parsed_text["timeLeft"])
        timer_processed = False
        if has_auth_timers:
            details_map = {
                NotificationType.STRUCTURE_LOST_SHIELD: gettext("Armor timer"),
                NotificationType.STRUCTURE_LOST_ARMOR: gettext("Final timer"),
            }
            AuthTimer.objects.create(
                details=details_map.get(self.notif_type, ""),
                system=structure_obj.eve_solar_system.name,
                planet_moon="",
                structure=structure_obj.eve_type.name,
                objective="Friendly",
                eve_time=eve_time,
                eve_corp=self.owner.corporation,
                corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
            )
            timer_processed = True

        if has_structure_timers:
            timer_map = {
                NotificationType.STRUCTURE_LOST_SHIELD: Timer.Type.ARMOR,
                NotificationType.STRUCTURE_LOST_ARMOR: Timer.Type.HULL,
            }
            eve_solar_system, _ = EveSolarSystem2.objects.get_or_create_esi(
                id=structure_obj.eve_solar_system_id
            )
            structure_type, _ = EveType2.objects.get_or_create_esi(
                id=structure_obj.eve_type_id
            )
            visibility = (
                Timer.Visibility.CORPORATION
                if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
                else Timer.Visibility.UNRESTRICTED
            )
            Timer.objects.create(
                eve_solar_system=eve_solar_system,
                structure_type=structure_type,
                timer_type=timer_map.get(self.notif_type),
                objective=Timer.Objective.FRIENDLY,
                date=eve_time,
                eve_corporation=self.owner.corporation,
                eve_alliance=self.owner.corporation.alliance,
                visibility=visibility,
                structure_name=structure_obj.name,
                owner_name=self.owner.corporation.corporation_name,
                details_notes=self._timer_details_notes(),
            )
            timer_processed = True

        return timer_processed

    def _gen_timer_sov_reinforcements(self, parsed_text: str) -> bool:
        """Generate timer for sov reinforcements."""
        if not self.owner.is_alliance_main:
            return False

        solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            id=parsed_text["solarSystemID"]
        )
        event_type = parsed_text["campaignEventType"]
        if event_type in self.MAP_CAMPAIGN_EVENT_2_TYPE_ID:
            structure_type_name = self.MAP_TYPE_ID_2_TIMER_STRUCTURE_NAME[
                self.type_id_from_event_type(parsed_text["campaignEventType"])
            ]
        else:
            structure_type_name = "Other"

        eve_time = ldap_time_2_datetime(parsed_text["decloakTime"])
        timer_processed = False
        if has_auth_timers:
            AuthTimer.objects.create(
                details=gettext("Sov timer"),
                system=solar_system.name,
                planet_moon="",
                structure=structure_type_name,
                objective="Friendly",
                eve_time=eve_time,
                eve_corp=self.owner.corporation,
                corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
            )
            timer_processed = True

        if has_structure_timers:
            eve_solar_system, _ = EveSolarSystem2.objects.get_or_create_esi(
                id=parsed_text["solarSystemID"]
            )
            structure_type, _ = EveType2.objects.get_or_create_esi(
                id=self.type_id_from_event_type(parsed_text["campaignEventType"])
            )
            visibility = (
                Timer.Visibility.CORPORATION
                if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
                else Timer.Visibility.UNRESTRICTED
            )
            Timer.objects.create(
                eve_solar_system=eve_solar_system,
                structure_type=structure_type,
                timer_type=Timer.Type.FINAL,
                objective=Timer.Objective.FRIENDLY,
                date=eve_time,
                eve_corporation=self.owner.corporation,
                eve_alliance=self.owner.corporation.alliance,
                visibility=visibility,
                owner_name=self.sender.name,
                details_notes=self._timer_details_notes(),
            )
            timer_processed = True

        return timer_processed

    def _gen_timer_orbital_reinforcements(self, parsed_text: str) -> bool:
        """Generate timer for orbital reinforcements."""
        solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            id=parsed_text["solarSystemID"]
        )
        planet, _ = EvePlanet.objects.get_or_create_esi(id=parsed_text["planetID"])
        eve_time = ldap_time_2_datetime(parsed_text["reinforceExitTime"])
        timer_processed = False
        if has_auth_timers:
            AuthTimer.objects.create(
                details=gettext("Final timer"),
                system=solar_system.name,
                planet_moon=planet.name,
                structure="POCO",
                objective="Friendly",
                eve_time=eve_time,
                eve_corp=self.owner.corporation,
                corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
            )
            timer_processed = True

        if has_structure_timers:
            eve_solar_system, _ = EveSolarSystem2.objects.get_or_create_esi(
                id=parsed_text["solarSystemID"]
            )
            structure_type, _ = EveType2.objects.get_or_create_esi(
                id=EveTypeId.CUSTOMS_OFFICE
            )
            visibility = (
                Timer.Visibility.CORPORATION
                if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
                else Timer.Visibility.UNRESTRICTED
            )
            Timer.objects.create(
                eve_solar_system=eve_solar_system,
                structure_type=structure_type,
                timer_type=Timer.Type.FINAL,
                objective=Timer.Objective.FRIENDLY,
                date=eve_time,
                location_details=planet.name,
                eve_corporation=self.owner.corporation,
                eve_alliance=self.owner.corporation.alliance,
                visibility=visibility,
                structure_name="Customs Office",
                owner_name=self.owner.corporation.corporation_name,
                details_notes=self._timer_details_notes(),
            )
            timer_processed = True

        return timer_processed

    def _gen_timer_moon_extraction(self, parsed_text: str) -> bool:
        """Generate timer for moon mining extractions."""
        solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            id=parsed_text["solarSystemID"]
        )
        moon, _ = EveMoon.objects.get_or_create_esi(id=parsed_text["moonID"])
        if "readyTime" in parsed_text:
            eve_time = ldap_time_2_datetime(parsed_text["readyTime"])
        else:
            eve_time = None
        details = gettext("Extraction ready")
        system = solar_system.name
        planet_moon = moon.name
        structure_type_name = "Moon Mining Cycle"
        objective = "Friendly"
        timer_processed = False

        if has_structure_timers:
            eve_solar_system, _ = EveSolarSystem2.objects.get_or_create_esi(
                id=parsed_text["solarSystemID"]
            )
            structure_type, _ = EveType2.objects.get_or_create_esi(
                id=parsed_text["structureTypeID"]
            )
            visibility = (
                Timer.Visibility.CORPORATION
                if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
                else Timer.Visibility.UNRESTRICTED
            )
        else:
            eve_solar_system = None
            structure_type = None
            visibility = None

        if self.notif_type == NotificationType.MOONMINING_EXTRACTION_STARTED:
            if has_auth_timers:
                AuthTimer.objects.create(
                    details=details,
                    system=system,
                    planet_moon=planet_moon,
                    structure=structure_type_name,
                    objective=objective,
                    eve_time=eve_time,
                    eve_corp=self.owner.corporation,
                    corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
                )
                timer_processed = True

            if has_structure_timers:
                eve_solar_system, _ = EveSolarSystem2.objects.get_or_create_esi(
                    id=parsed_text["solarSystemID"]
                )
                structure_type, _ = EveType2.objects.get_or_create_esi(
                    id=parsed_text["structureTypeID"]
                )
                visibility = (
                    Timer.Visibility.CORPORATION
                    if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
                    else Timer.Visibility.UNRESTRICTED
                )
                Timer.objects.create(
                    eve_solar_system=eve_solar_system,
                    structure_type=structure_type,
                    timer_type=Timer.Type.MOONMINING,
                    objective=Timer.Objective.FRIENDLY,
                    date=eve_time,
                    location_details=moon.name,
                    eve_corporation=self.owner.corporation,
                    eve_alliance=self.owner.corporation.alliance,
                    visibility=visibility,
                    structure_name=parsed_text["structureName"],
                    owner_name=self.owner.corporation.corporation_name,
                    details_notes=self._timer_details_notes(),
                )
                timer_processed = True

        elif self.notif_type == NotificationType.MOONMINING_EXTRACTION_CANCELLED:
            timer_processed = True
            notifications_qs = Notification.objects.filter(
                notif_type=NotificationType.MOONMINING_EXTRACTION_STARTED,
                owner=self.owner,
                is_timer_added=True,
                timestamp__lte=self.timestamp,
            ).order_by("-timestamp")
            for notification in notifications_qs:
                parsed_text_2 = notification.parsed_text()
                my_structure_type_id = parsed_text_2["structureTypeID"]
                if my_structure_type_id == parsed_text["structureTypeID"]:
                    eve_time = ldap_time_2_datetime(parsed_text_2["readyTime"])
                    if has_auth_timers:
                        timer_query = AuthTimer.objects.filter(
                            system=system,
                            planet_moon=planet_moon,
                            structure=structure_type_name,
                            objective=objective,
                            eve_time=eve_time,
                        )
                        deleted_count, _ = timer_query.delete()
                        logger.info(
                            f"{self.notification_id}: removed {deleted_count} "
                            "obsolete Auth timers related to notification"
                        )

                    if has_structure_timers:
                        timer_query = Timer.objects.filter(
                            eve_solar_system=eve_solar_system,
                            structure_type=structure_type,
                            timer_type=Timer.Type.MOONMINING,
                            location_details=moon.name,
                            date=eve_time,
                            objective=Timer.Objective.FRIENDLY,
                            eve_corporation=self.owner.corporation,
                            eve_alliance=self.owner.corporation.alliance,
                            visibility=visibility,
                            structure_name=parsed_text["structureName"],
                            owner_name=self.owner.corporation.corporation_name,
                        )
                        deleted_count, _ = timer_query.delete()
                        logger.info(
                            f"{self.notification_id}: removed {deleted_count} "
                            "obsolete structure timers related to notification"
                        )

        return timer_processed

    def _gen_timer_tower_reinforcements(self, parsed_text: str) -> bool:
        """Generate timer for tower reinforcements."""
        structure = self.structures.first()
        eve_time = dt.datetime.fromisoformat(self.details["reinforced_until"])
        timer_processed = False
        if has_auth_timers:
            structure_type_map = {
                starbases.StarbaseSize.SMALL: "POS[S]",
                starbases.StarbaseSize.MEDIUM: "POS[M]",
                starbases.StarbaseSize.LARGE: "POS[L]",
            }
            structure_str = structure_type_map.get(
                starbases.starbase_size(structure.eve_type), "POS[M]"
            )
            AuthTimer.objects.create(
                details=gettext("Final timer"),
                system=structure.eve_solar_system.name,
                planet_moon=structure.eve_moon.name,
                structure=structure_str,
                objective="Friendly",
                eve_time=eve_time,
                eve_corp=self.owner.corporation,
                corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
            )
            timer_processed = True

        if has_structure_timers:
            eve_solar_system, _ = EveSolarSystem2.objects.get_or_create_esi(
                id=structure.eve_solar_system.id
            )
            structure_type, _ = EveType2.objects.get_or_create_esi(
                id=structure.eve_type.id
            )
            visibility = (
                Timer.Visibility.CORPORATION
                if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
                else Timer.Visibility.UNRESTRICTED
            )
            Timer.objects.create(
                eve_solar_system=eve_solar_system,
                structure_type=structure_type,
                timer_type=Timer.Type.FINAL,
                objective=Timer.Objective.FRIENDLY,
                date=eve_time,
                location_details=structure.eve_moon.name,
                eve_corporation=self.owner.corporation,
                eve_alliance=self.owner.corporation.alliance,
                visibility=visibility,
                structure_name=structure.name,
                owner_name=self.owner.corporation.corporation_name,
                details_notes=self._timer_details_notes(),
            )
            timer_processed = True

        return timer_processed

    def _timer_details_notes(self) -> str:
        """Return generated details notes string for Timers."""
        return (
            "Automatically created from structure notification for "
            f"{self.owner.corporation} at {self.timestamp.strftime(DATETIME_FORMAT)}"
        )

    @classmethod
    def type_id_from_event_type(cls, event_type: int) -> Optional[int]:
        if event_type in cls.MAP_CAMPAIGN_EVENT_2_TYPE_ID:
            return cls.MAP_CAMPAIGN_EVENT_2_TYPE_ID[event_type]
        return None


class Notification(NotificationBase):
    """A notification in Eve Online.

    Notifications are usually created from Eve Online notifications received from ESI,
    but they can also be generated directly by Structures.
    """

    TEMPORARY_NOTIFICATION_ID = 999999999999

    notification_id = models.PositiveBigIntegerField(verbose_name="id")
    created = models.DateTimeField(
        null=True,
        default=None,
        help_text="Date when this notification was first received from ESI",
    )
    is_read = models.BooleanField(
        null=True,
        default=None,
        help_text="True when this notification has read in the eve client",
    )
    last_updated = models.DateTimeField(
        help_text="Date when this notification has last been updated from ESI"
    )
    sender = models.ForeignKey(EveEntity, on_delete=models.SET_NULL, null=True)

    text = models.TextField(
        null=True, default=None, blank=True, help_text="Notification details in YAML"
    )
    timestamp = models.DateTimeField(db_index=True)

    objects = NotificationManager()

    class Meta:
        verbose_name = "eve notification"
        unique_together = (("notification_id", "owner"),)

    def save(self, *args, **kwargs) -> None:
        if self.is_temporary:
            raise ValueError("Temporary notifications can not be saved")
        super().save(*args, **kwargs)

    @property
    def is_temporary(self) -> bool:
        """True when this notification is temporary."""
        return self.notification_id == self.TEMPORARY_NOTIFICATION_ID

    @property
    def is_generated(self) -> bool:
        return self.is_temporary

    # @classmethod
    # def get_all_types(cls) -> Set[int]:
    #     """returns a set with all supported notification types"""
    #     return {x[0] for x in NotificationType.choices}

    def parsed_text(self) -> dict:
        """Returns the notifications's text as dict."""
        return yaml.safe_load(self.text)

    def is_npc_attacking(self) -> bool:
        """Whether this notification is about a NPC attacking."""
        if self.notif_type in [
            NotificationType.ORBITAL_ATTACKED,
            NotificationType.STRUCTURE_UNDER_ATTACK,
        ]:
            parsed_text = self.parsed_text()
            corporation_id = None
            if self.notif_type == NotificationType.STRUCTURE_UNDER_ATTACK:
                if (
                    "corpLinkData" in parsed_text
                    and len(parsed_text["corpLinkData"]) >= 3
                ):
                    corporation_id = int(parsed_text["corpLinkData"][2])
            if self.notif_type == NotificationType.ORBITAL_ATTACKED:
                if "aggressorCorpID" in parsed_text:
                    corporation_id = int(parsed_text["aggressorCorpID"])
            if corporation_id:
                corporation = EveEntity2(
                    category=EveEntity2.CATEGORY_CORPORATION, id=corporation_id
                )
                return corporation.is_npc and not corporation.is_npc_starter_corporation
        return False

    def update_related_structures(self) -> bool:
        """Update related structure for this notification.

        Returns True if structures where updated, else False.
        """
        structures_qs = self.calc_related_structures()
        self.structures.clear()
        if structures_qs.exists():
            objs = [obj for obj in structures_qs.all()]
            self.structures.add(*objs)
            return True
        return False

    @classmethod
    def create_from_structure(
        cls, structure: Structure, notif_type: NotificationType, **kwargs
    ) -> "Notification":
        """Create new notification from given structure."""
        if "timestamp" not in kwargs:
            kwargs["timestamp"] = now()
        if "last_updated" not in kwargs:
            kwargs["last_updated"] = now()
        threshold = kwargs.pop("threshold") if "threshold" in kwargs else None
        if "text" not in kwargs:
            if notif_type in {
                NotificationType.STRUCTURE_FUEL_ALERT,
                NotificationType.STRUCTURE_REFUELED_EXTRA,
                NotificationType.STRUCTURE_JUMP_FUEL_ALERT,
            }:
                data = {
                    "solarsystemID": structure.eve_solar_system_id,
                    "structureID": structure.id,
                    "structureTypeID": structure.eve_type_id,
                    "threshold": threshold,
                }
            elif notif_type in {
                NotificationType.TOWER_RESOURCE_ALERT_MSG,
                NotificationType.TOWER_REFUELED_EXTRA,
            }:
                data = {
                    "moonID": structure.eve_moon_id,
                    "typeID": structure.eve_type_id,
                    "structureID": structure.id,
                }
            else:
                raise ValueError("text property not provided and can not be generated.")
            kwargs["text"] = yaml.dump(data)
        if "sender" not in kwargs:
            sender, _ = EveEntity.objects.get_or_create_esi(id=EveCorporationId.DED)
            kwargs["sender"] = sender
        kwargs["notification_id"] = cls.TEMPORARY_NOTIFICATION_ID
        kwargs["owner"] = structure.owner
        kwargs["notif_type"] = notif_type
        return cls(**kwargs)


class GeneratedNotification(NotificationBase):
    """A notification generated by the Structures app, not by Eve Online."""

    details = models.JSONField(default=dict)
    last_updated = models.DateTimeField(auto_now=True)
    timestamp = models.DateTimeField(auto_now_add=True)

    objects = GeneratedNotificationManager()

    @property
    def notification_id(self) -> int:
        return self.pk

    @property
    def is_temporary(self) -> bool:
        return False

    @property
    def is_generated(self) -> bool:
        return True

    def is_npc_attacking(self) -> bool:
        return False

    def parsed_text(self) -> dict:
        """Adopting to Notification API."""
        return self.details


class BaseFuelAlertConfig(models.Model):
    """Configuration of structure fuel notifications."""

    channel_ping_type = models.CharField(
        max_length=2,
        choices=Webhook.PingType.choices,
        default=Webhook.PingType.HERE,
        verbose_name="channel pings",
        help_text=(
            "Option to ping every member of the channel. "
            "This setting can be overruled by the respective owner "
            "or webhook configuration"
        ),
    )
    color = models.IntegerField(
        choices=Webhook.Color.choices,
        default=Webhook.Color.WARNING,
        blank=True,
        null=True,
        help_text="Context color of these notification on Discord",
    )
    is_enabled = models.BooleanField(
        default=True,
        help_text="Disabled configurations will not create any new alerts.",
    )

    class Meta:
        abstract = True

    def __str__(self) -> str:
        return f"#{self.pk}"

    def send_new_notifications(self, force: bool = False) -> None:
        """Send new fuel notifications based on this config."""
        raise NotImplementedError()

    @staticmethod
    def relevant_webhooks() -> models.QuerySet:
        """Webhooks relevant for processing fuel notifications based on this config."""
        raise NotImplementedError()


class FuelAlertConfig(BaseFuelAlertConfig):
    """Configuration of structure fuel notifications."""

    end = models.PositiveIntegerField(
        help_text="End of alerts in hours before fuel expires"
    )
    repeat = models.PositiveIntegerField(
        help_text=(
            "Notifications will be repeated every x hours. Set to 0 for no repeats"
        )
    )
    start = models.PositiveIntegerField(
        help_text="Start of alerts in hours before fuel expires"
    )

    class Meta:
        verbose_name = "structure fuel alert config"

    def clean(self) -> None:
        if self.start <= self.end:
            raise ValidationError(
                _("Start must be before end, i.e. have a larger value.")
            )
        if self.repeat >= self.start - self.end:
            raise ValidationError(
                {"repeat": _("Repeat can not be larger that the interval size.")}
            )
        new = range(self.end, self.start)
        for config in FuelAlertConfig.objects.exclude(pk=self.pk):
            current = range(config.end, config.start)
            overlap = range(max(new[0], current[0]), min(new[-1], current[-1]) + 1)
            if len(overlap) > 0:
                raise ValidationError(
                    _(
                        "This configuration may not overlap with an "
                        "existing configuration."
                    )
                )

    def save(self, *args, **kwargs) -> None:
        try:
            old_instance = FuelAlertConfig.objects.get(pk=self.pk)
        except FuelAlertConfig.DoesNotExist:
            old_instance = None
        super().save(*args, **kwargs)
        if old_instance and (
            old_instance.start != self.start
            or old_instance.end != self.end
            or old_instance.repeat != self.repeat
        ):
            self.structure_fuel_alerts.all().delete()

    def send_new_notifications(self, force: bool = False) -> None:
        """Send new fuel notifications based on this config."""
        structures = Structure.objects.filter(
            eve_type__eve_group__eve_category_id__in=[
                EveCategoryId.STARBASE,
                EveCategoryId.STRUCTURE,
            ],
            fuel_expires_at__isnull=False,
        )
        for structure in structures:
            if not structure.is_burning_fuel:
                continue
            hours_left = structure.hours_fuel_expires
            if self.start >= hours_left >= self.end:
                hours_last_alert = (
                    self.start
                    - (
                        math.floor((self.start - hours_left) / self.repeat)
                        * self.repeat
                    )
                    if self.repeat
                    else self.start
                )
                notif, created = FuelAlert.objects.get_or_create(
                    structure=structure, config=self, hours=hours_last_alert
                )
                if created or force:
                    notif.send_generated_notification()

    @staticmethod
    def relevant_webhooks() -> models.QuerySet:
        """Webhooks relevant for processing fuel notifications based on this config."""
        return (
            Webhook.objects.filter(is_active=True)
            .filter(Q(owners__isnull=False) | Q(structures__isnull=False))
            .filter(
                Q(notification_types__contains=NotificationType.STRUCTURE_FUEL_ALERT)
                | Q(
                    notification_types__contains=NotificationType.TOWER_RESOURCE_ALERT_MSG
                )
            )
            .distinct()
        )


class JumpFuelAlertConfig(BaseFuelAlertConfig):
    """Configuration of jump fuel notifications."""

    threshold = models.PositiveIntegerField(
        help_text=(
            "Notifications will be sent once fuel level in units reaches this threshold"
        )
    )

    def save(self, *args, **kwargs) -> None:
        try:
            old_instance = JumpFuelAlertConfig.objects.get(pk=self.pk)
        except JumpFuelAlertConfig.DoesNotExist:
            old_instance = None
        super().save(*args, **kwargs)
        if old_instance and (old_instance.threshold != self.threshold):
            self.jump_fuel_alerts.all().delete()

    def send_new_notifications(self, force: bool = False) -> None:
        """Send new fuel notifications based on this config."""
        jump_gates = Structure.objects.filter(eve_type_id=EveTypeId.JUMP_GATE)
        for jump_gate in jump_gates:
            if not jump_gate.is_burning_fuel:
                continue
            fuel_quantity = jump_gate.jump_fuel_quantity()
            if fuel_quantity and fuel_quantity < self.threshold:
                notif, created = JumpFuelAlert.objects.get_or_create(
                    structure=jump_gate, config=self
                )
                if created or force:
                    notif.send_generated_notification()

    @staticmethod
    def relevant_webhooks() -> models.QuerySet:
        """Webhooks relevant for processing jump fuel notifications based on this config."""
        return Webhook.objects.filter(
            is_active=True,
            notification_types__contains=NotificationType.STRUCTURE_JUMP_FUEL_ALERT,
        ).filter(Q(owners__isnull=False) | Q(structures__isnull=False))


class BaseFuelAlert(models.Model):
    class Meta:
        abstract = True

    def send_generated_notification(self):
        raise NotImplementedError()


class FuelAlert(BaseFuelAlert):
    """A generated notification alerting about fuel getting low in structures."""

    structure = models.ForeignKey(
        Structure, on_delete=models.CASCADE, related_name="structure_fuel_alerts"
    )
    config = models.ForeignKey(
        FuelAlertConfig,
        on_delete=models.CASCADE,
        related_name="structure_fuel_alerts",
    )
    hours = models.PositiveIntegerField(
        db_index=True,
        help_text="number of hours before fuel expiration this alert was sent",
    )

    class Meta:
        verbose_name = "structure fuel alert"
        constraints = [
            models.UniqueConstraint(
                fields=["structure", "config", "hours"],
                name="functional_pk_fuelalert",
            )
        ]

    def __str__(self) -> str:
        return f"{self.structure}-{self.config}-{self.hours}"

    def send_generated_notification(self):
        notif_type = (
            NotificationType.TOWER_RESOURCE_ALERT_MSG
            if self.structure.is_starbase
            else NotificationType.STRUCTURE_FUEL_ALERT
        )
        notif = Notification.create_from_structure(
            structure=self.structure, notif_type=notif_type
        )
        notif.send_to_configured_webhooks(
            ping_type_override=self.config.channel_ping_type,
            use_color_override=True,
            color_override=self.config.color,
        )


class JumpFuelAlert(BaseFuelAlert):
    """A generated notification alerting about jump fuel getting low."""

    structure = models.ForeignKey(
        Structure, on_delete=models.CASCADE, related_name="jump_fuel_alerts"
    )
    config = models.ForeignKey(
        JumpFuelAlertConfig, on_delete=models.CASCADE, related_name="jump_fuel_alerts"
    )

    def __str__(self) -> str:
        return f"{self.structure}-{self.config}"

    def send_generated_notification(self) -> None:
        notif = Notification.create_from_structure(
            structure=self.structure,
            notif_type=NotificationType.STRUCTURE_JUMP_FUEL_ALERT,
            threshold=self.config.threshold,
        )
        notif.send_to_configured_webhooks(
            ping_type_override=self.config.channel_ping_type,
            use_color_override=True,
            color_override=self.config.color,
        )



# Generated by Django 4.0.3 on 2022-05-30 20:48

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0037_add_billing_notification_types"),
    ]

    operations = [
        migrations.CreateModel(
            name="StarbaseDetail",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("allow_alliance_members", models.BooleanField()),
                ("allow_corporation_members", models.BooleanField()),
                (
                    "anchor_role",
                    models.CharField(
                        choices=[
                            ("AL", "alliance member"),
                            ("CE", "config starbase equipment role"),
                            ("CO", "corporation member"),
                            ("FT", "starbase fuel technician role"),
                        ],
                        max_length=2,
                    ),
                ),
                ("attack_if_at_war", models.BooleanField()),
                ("attack_if_other_security_status_dropping", models.BooleanField()),
                (
                    "attack_security_status_threshold",
                    models.FloatField(default=None, null=True),
                ),
                (
                    "attack_standing_threshold",
                    models.FloatField(default=None, null=True),
                ),
                (
                    "fuel_bay_take_role",
                    models.CharField(
                        choices=[
                            ("AL", "alliance member"),
                            ("CE", "config starbase equipment role"),
                            ("CO", "corporation member"),
                            ("FT", "starbase fuel technician role"),
                        ],
                        max_length=2,
                    ),
                ),
                (
                    "fuel_bay_view_role",
                    models.CharField(
                        choices=[
                            ("AL", "alliance member"),
                            ("CE", "config starbase equipment role"),
                            ("CO", "corporation member"),
                            ("FT", "starbase fuel technician role"),
                        ],
                        max_length=2,
                    ),
                ),
                (
                    "last_modified_at",
                    models.DateTimeField(
                        help_text="When data was modified on the server."
                    ),
                ),
                (
                    "offline_role",
                    models.CharField(
                        choices=[
                            ("AL", "alliance member"),
                            ("CE", "config starbase equipment role"),
                            ("CO", "corporation member"),
                            ("FT", "starbase fuel technician role"),
                        ],
                        max_length=2,
                    ),
                ),
                (
                    "online_role",
                    models.CharField(
                        choices=[
                            ("AL", "alliance member"),
                            ("CE", "config starbase equipment role"),
                            ("CO", "corporation member"),
                            ("FT", "starbase fuel technician role"),
                        ],
                        max_length=2,
                    ),
                ),
                (
                    "unanchor_role",
                    models.CharField(
                        choices=[
                            ("AL", "alliance member"),
                            ("CE", "config starbase equipment role"),
                            ("CO", "corporation member"),
                            ("FT", "starbase fuel technician role"),
                        ],
                        max_length=2,
                    ),
                ),
                ("use_alliance_standings", models.BooleanField()),
                (
                    "structure",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="starbase_detail",
                        to="structures.structure",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="StarbaseDetailFuel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("quantity", models.IntegerField()),
                (
                    "detail",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fuels",
                        to="structures.starbasedetail",
                    ),
                ),
                (
                    "eve_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="structures.evetype",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 4.0.7 on 2022-08-17 18:32

from django.db import migrations


def forwards(apps, schema_editor):
    Notification = apps.get_model("structures", "Notification")
    qs = Notification.objects.filter(sender__category=5)
    print(f"Updating {qs.count()} notifications with 'other' senders.")
    qs.update(sender=None)


def backwards(apps, schema_editor):
    EveEntity = apps.get_model("structures", "EveEntity")
    ded, _ = EveEntity.objects.get_or_create(
        id=1000137, defaults={"name": "DED", "category": 2}
    )
    Notification = apps.get_model("structures", "Notification")
    qs = Notification.objects.filter(sender__isnull=True)
    print(f"Patching {qs.count()} notifications without a sender.")
    qs.update(sender=ded)


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0041_notification_sender_can_be_null"),
    ]

    operations = [
        migrations.RunPython(forwards, backwards),
    ]



# Generated by Django 4.0.7 on 2022-08-17 17:15

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0040_add_generated_notifications"),
    ]

    operations = [
        migrations.AlterField(
            model_name="notification",
            name="sender",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="structures.eveentity",
            ),
        ),
    ]



from typing import Iterable, Optional

from celery import chain, shared_task

from django.contrib.auth.models import User

from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce
from app_utils.esi import fetch_esi_status
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import STRUCTURES_TASKS_TIME_LIMIT
from .models import (
    EveSovereigntyMap,
    FuelAlertConfig,
    JumpFuelAlertConfig,
    NotificationType,
    Owner,
    Webhook,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

TASK_PRIORITY_HIGH = 2


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def update_all_structures():
    """Update all structures.

    Main task for starting regular update of all structures
    and related data from ESI.
    """
    if not fetch_esi_status().is_ok:
        logger.warning("ESI currently not available. Aborting.")
        return
    chain(update_sov_map.si(), update_structures.si()).delay()


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def update_sov_map():
    """Update sovereignty map from ESI."""
    EveSovereigntyMap.objects.update_from_esi()


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def update_structures():
    """Update all structures for all active owners from ESI."""
    for owner in Owner.objects.all():
        if owner.is_active:
            update_structures_for_owner.delay(owner.pk)

    if (
        Owner.objects.filter(is_active=True).count() > 0
        and Owner.objects.filter(is_active=True, is_alliance_main=True).count() == 0
    ):
        logger.warning(
            "No owner configured to process alliance wide notifications. "
            "Please set 'is alliance main' to True for the designated owner."
        )


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def update_all_for_owner(owner_pk, user_pk=None):
    """Update structures and notifications for owner from ESI."""
    chain(
        update_structures_for_owner.si(owner_pk, user_pk),
        process_notifications_for_owner.si(owner_pk, user_pk),
    ).delay()


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def update_structures_for_owner(owner_pk, user_pk=None):
    """Fetch all structures for owner and update related corp assets from ESI."""
    if not fetch_esi_status().is_ok:
        logger.warning("ESI currently not available. Aborting.")
    else:
        chain(
            update_structures_esi_for_owner.si(owner_pk, user_pk),
            update_structures_assets_for_owner.si(owner_pk, user_pk),
        ).delay()


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def update_structures_esi_for_owner(owner_pk, user_pk=None):
    """Update all structures for owner for ESI."""
    owner = Owner.objects.get(pk=owner_pk)
    owner.update_structures_esi(_get_user(user_pk))


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def update_structures_assets_for_owner(owner_pk, user_pk=None):
    """Update all related assets for owner."""
    owner = Owner.objects.get(pk=owner_pk)
    owner.update_asset_esi(_get_user(user_pk))


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def fetch_all_notifications():
    """Fetch notifications for all owners and send new fuel notifications."""
    for owner in Owner.objects.all():
        if owner.is_active:
            owner.update_is_up()
            process_notifications_for_owner.apply_async(
                kwargs={"owner_pk": owner.pk}, priority=TASK_PRIORITY_HIGH
            )
    for config_pk in FuelAlertConfig.objects.filter(is_enabled=True).values_list(
        "pk", flat=True
    ):
        send_structure_fuel_notifications_for_config.delay(config_pk)
    for config_pk in JumpFuelAlertConfig.objects.filter(is_enabled=True).values_list(
        "pk", flat=True
    ):
        send_jump_fuel_notifications_for_config.delay(config_pk)


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def process_notifications_for_owner(owner_pk: int, user_pk: int = None):
    """Fetch all notification for owner from ESI and processes them."""
    if not fetch_esi_status().is_ok:
        logger.warning("ESI currently not available. Aborting.")
        return
    chain(
        fetch_notification_for_owner.si(owner_pk=owner_pk, user_pk=user_pk),
        update_existing_notifications.si(owner_pk=owner_pk),
        send_new_notifications_for_owner.si(owner_pk=owner_pk),
        generate_new_timers_for_owner.si(owner_pk=owner_pk),
    ).apply_async(priority=TASK_PRIORITY_HIGH)


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def fetch_notification_for_owner(owner_pk: int, user_pk: int = None):
    """Fetch notifications from ESI."""
    owner = Owner.objects.get(pk=owner_pk)
    owner.fetch_notifications_esi(_get_user(user_pk))


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def update_existing_notifications(owner_pk: int) -> int:
    """Update structure relation for existing notifications if needed.

    Returns number of updated notifications.
    """
    owner = Owner.objects.get(pk=owner_pk)
    notif_need_update_qs = owner.notification_set.filter(
        notif_type__in=NotificationType.structure_related, structures__isnull=True
    )
    notif_need_update_count = notif_need_update_qs.count()
    updated_count = 0
    if notif_need_update_count > 0:
        for notif in notif_need_update_qs:
            if notif.update_related_structures():
                updated_count += 1
        logger.info(
            "%s: Updated structure relation for %d of %d relevant notifications",
            owner,
            updated_count,
            notif_need_update_count,
        )
    return updated_count


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def generate_new_timers_for_owner(owner_pk: int):
    owner = Owner.objects.get(pk=owner_pk)
    owner.add_or_remove_timers_from_notifications()


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def send_new_notifications_for_owner(owner_pk: int):
    """Send new notifications to Discord."""
    owner = Owner.objects.get(pk=owner_pk)
    owner.send_new_notifications()
    send_queued_messages_for_webhooks(owner.webhooks.filter(is_active=True))


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def send_structure_fuel_notifications_for_config(config_pk: int):
    FuelAlertConfig.objects.get(pk=config_pk).send_new_notifications()
    send_queued_messages_for_webhooks(FuelAlertConfig.relevant_webhooks())


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def send_jump_fuel_notifications_for_config(config_pk: int):
    JumpFuelAlertConfig.objects.get(pk=config_pk).send_new_notifications()
    send_queued_messages_for_webhooks(JumpFuelAlertConfig.relevant_webhooks())


def send_queued_messages_for_webhooks(webhooks: Iterable[Webhook]):
    """Send queued message for given webhooks."""
    for webhook in webhooks:
        if webhook.queue_size() > 0:
            send_messages_for_webhook.apply_async(
                kwargs={"webhook_pk": webhook.pk}, priority=TASK_PRIORITY_HIGH
            )


@shared_task(base=QueueOnce)
def send_messages_for_webhook(webhook_pk: int) -> None:
    """Send all currently queued messages for given webhook to Discord."""
    Webhook.objects.send_queued_messages_for_webhook(webhook_pk)


@shared_task(time_limit=STRUCTURES_TASKS_TIME_LIMIT)
def send_test_notifications_to_webhook(webhook_pk, user_pk=None) -> None:
    """Send test notification to given webhook."""
    webhook = Webhook.objects.get(pk=webhook_pk)
    user = _get_user(user_pk)
    send_report, send_success = webhook.send_test_message(user)
    if user:
        message = 'Test notification to webhook "{}" {}.\n'.format(
            webhook, "completed successfully" if send_success else "has failed"
        )
        if not send_success:
            message += "Error: {}".format(send_report)
        notify(
            user=user,
            title='{}: Test notification to "{}": {}'.format(
                __title__, webhook, "OK" if send_success else "FAILED"
            ),
            message=message,
            level="success" if send_success else "danger",
        )


def _get_user(user_pk: int) -> Optional[User]:
    """Fetch the user or return None."""
    if not user_pk:
        return None
    try:
        return User.objects.get(pk=user_pk)
    except User.DoesNotExist:
        logger.warning("Ignoring non-existing user with pk %s", user_pk)
        return None



import math
from enum import IntEnum, auto
from typing import Optional

from ..constants import EveGroupId
from ..models.eveuniverse import EveType


class StarbaseSize(IntEnum):
    SMALL = auto()
    MEDIUM = auto()
    LARGE = auto()


def is_starbase(eve_type: EveType) -> bool:
    return eve_type.eve_group_id == EveGroupId.CONTROL_TOWER


def starbase_size(eve_type: EveType) -> StarbaseSize:
    """return the size of a starbase or None if this type is not a starbase"""
    if not is_starbase(eve_type):
        return None
    elif "medium" in eve_type.name.lower():
        return StarbaseSize.MEDIUM
    elif "small" in eve_type.name.lower():
        return StarbaseSize.SMALL
    return StarbaseSize.LARGE


def fuel_per_hour(eve_type: EveType) -> Optional[int]:
    """Calculate the number of fuel blocks consumed per hour.

    Returns None if not a starbase.
    """
    size = starbase_size(eve_type)
    if size is StarbaseSize.LARGE:
        return 40
    elif size is StarbaseSize.MEDIUM:
        return 20
    elif size is StarbaseSize.SMALL:
        return 10
    return None


def fuel_duration(
    starbase_type: EveType,
    fuel_quantity: int,
    has_sov: bool = False,
) -> float:
    """Calculate how long the fuel lasts in seconds."""
    sov_discount = 0.25 if has_sov else 0
    amount_per_hour = fuel_per_hour(starbase_type)
    if amount_per_hour is None:
        raise ValueError(
            f"{starbase_type}: Can only calculate fuel durations for starbases"
        )
    seconds = math.floor(3600 * fuel_quantity / (amount_per_hour * (1 - sov_discount)))
    return seconds



# Generated by Django 3.1.6 on 2021-04-30 14:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0023_structure_fittings"),
    ]

    operations = [
        migrations.AlterField(
            model_name="evecategory",
            name="name",
            field=models.CharField(help_text="Eve Online name", max_length=255),
        ),
        migrations.AlterField(
            model_name="evecategory",
            name="name_de",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for German",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evecategory",
            name="name_ko",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Korean",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evecategory",
            name="name_ru",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Russian",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evecategory",
            name="name_zh",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Chinese",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="eveconstellation",
            name="name",
            field=models.CharField(help_text="Eve Online name", max_length=255),
        ),
        migrations.AlterField(
            model_name="eveconstellation",
            name="name_de",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for German",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="eveconstellation",
            name="name_ko",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Korean",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="eveconstellation",
            name="name_ru",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Russian",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="eveconstellation",
            name="name_zh",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Chinese",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evegroup",
            name="name",
            field=models.CharField(help_text="Eve Online name", max_length=255),
        ),
        migrations.AlterField(
            model_name="evegroup",
            name="name_de",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for German",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evegroup",
            name="name_ko",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Korean",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evegroup",
            name="name_ru",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Russian",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evegroup",
            name="name_zh",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Chinese",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evemoon",
            name="name",
            field=models.CharField(help_text="Eve Online name", max_length=255),
        ),
        migrations.AlterField(
            model_name="evemoon",
            name="name_de",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for German",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evemoon",
            name="name_ko",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Korean",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evemoon",
            name="name_ru",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Russian",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evemoon",
            name="name_zh",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Chinese",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="eveplanet",
            name="name",
            field=models.CharField(help_text="Eve Online name", max_length=255),
        ),
        migrations.AlterField(
            model_name="eveplanet",
            name="name_de",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for German",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="eveplanet",
            name="name_ko",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Korean",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="eveplanet",
            name="name_ru",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Russian",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="eveplanet",
            name="name_zh",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Chinese",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="everegion",
            name="name",
            field=models.CharField(help_text="Eve Online name", max_length=255),
        ),
        migrations.AlterField(
            model_name="everegion",
            name="name_de",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for German",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="everegion",
            name="name_ko",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Korean",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="everegion",
            name="name_ru",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Russian",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="everegion",
            name="name_zh",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Chinese",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evesolarsystem",
            name="name",
            field=models.CharField(help_text="Eve Online name", max_length=255),
        ),
        migrations.AlterField(
            model_name="evesolarsystem",
            name="name_de",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for German",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evesolarsystem",
            name="name_ko",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Korean",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evesolarsystem",
            name="name_ru",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Russian",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evesolarsystem",
            name="name_zh",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Chinese",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evetype",
            name="name",
            field=models.CharField(help_text="Eve Online name", max_length=255),
        ),
        migrations.AlterField(
            model_name="evetype",
            name="name_de",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for German",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evetype",
            name="name_ko",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Korean",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evetype",
            name="name_ru",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Russian",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="evetype",
            name="name_zh",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Chinese",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="structureservice",
            name="name_de",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for German",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="structureservice",
            name="name_ko",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Korean",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="structureservice",
            name="name_ru",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Russian",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="structureservice",
            name="name_zh",
            field=models.CharField(
                blank=True,
                help_text="Eve Online name localized for Chinese",
                max_length=255,
            ),
        ),
    ]



# Generated by Django 2.2.13 on 2020-08-24 13:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("auth", "0011_update_proxy_permissions"),
        ("structures", "0012_disable_pings"),
    ]

    operations = [
        migrations.RenameField(
            model_name="owner",
            old_name="has_pings_enabled",
            new_name="has_default_pings_enabled",
        ),
        migrations.RenameField(
            model_name="webhook",
            old_name="has_pings_enabled",
            new_name="has_default_pings_enabled",
        ),
        migrations.RemoveField(
            model_name="structure",
            name="next_reinforce_weekday",
        ),
        migrations.RemoveField(
            model_name="structure",
            name="reinforce_weekday",
        ),
        migrations.AddField(
            model_name="owner",
            name="ping_groups",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Groups to be pinged for each notification - ",
                to="auth.Group",
            ),
        ),
        migrations.AddField(
            model_name="webhook",
            name="ping_groups",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Groups to be pinged for each notification - ",
                to="auth.Group",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="next_reinforce_apply",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="The requested change to reinforce_hour that will take effect at the time shown by next_reinforce_apply",
                null=True,
            ),
        ),
    ]



import functools
import urllib
from collections import defaultdict
from enum import IntEnum

from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.staticfiles.storage import staticfiles_storage
from django.db.models import Count, Q
from django.http import HttpResponse, HttpResponseServerError, JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.utils import translation
from django.utils.html import format_html
from django.utils.translation import gettext as _
from esi.decorators import token_required
from eveuniverse.core import eveimageserver
from eveuniverse.models import EveType as EveType2
from eveuniverse.models import EveTypeDogmaAttribute

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from app_utils.allianceauth import notify_admins
from app_utils.logging import LoggerAddTag
from app_utils.messages import messages_plus
from app_utils.views import image_html

from . import __title__, tasks
from .app_settings import (
    STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED,
    STRUCTURES_DEFAULT_LANGUAGE,
    STRUCTURES_DEFAULT_PAGE_LENGTH,
    STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED,
    STRUCTURES_PAGING_ENABLED,
    STRUCTURES_SHOW_JUMP_GATES,
)
from .constants import EveAttributeId, EveCategoryId, EveGroupId, EveTypeId
from .core.serializers import (
    JumpGatesListSerializer,
    PocoListSerializer,
    StructureListSerializer,
)
from .forms import TagsFilterForm
from .models import EveType, Owner, Structure, StructureItem, StructureTag, Webhook

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

QUERY_PARAM_TAGS = "tags"


def default_if_none(value, default=None):
    """Return default if a value is None."""
    if value is None:
        return default
    return value


@login_required
@permission_required("structures.basic_access")
def index(request):
    url = reverse("structures:main")
    if STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED:
        params = {
            QUERY_PARAM_TAGS: ",".join(
                [x.name for x in StructureTag.objects.filter(is_default=True)]
            )
        }
        url += "?{}".format(urllib.parse.urlencode(params))
    return redirect(url)


@login_required
@permission_required("structures.basic_access")
def main(request):
    """Main view"""
    active_tags = list()
    if request.method == "POST":
        form = TagsFilterForm(data=request.POST)
        if form.is_valid():
            for name, activated in form.cleaned_data.items():
                if activated:
                    active_tags.append(get_object_or_404(StructureTag, name=name))

            url = reverse("structures:main")
            if active_tags:
                params = {QUERY_PARAM_TAGS: ",".join([x.name for x in active_tags])}
                url += "?{}".format(urllib.parse.urlencode(params))
            return redirect(url)
    else:
        tags_raw = request.GET.get(QUERY_PARAM_TAGS)
        if tags_raw:
            tags_parsed = tags_raw.split(",")
            active_tags = [
                x for x in StructureTag.objects.all() if x.name in tags_parsed
            ]
        form = TagsFilterForm(initial={x.name: True for x in active_tags})

    context = {
        "page_title": _(__title__),
        "active_tags": active_tags,
        "tags_filter_form": form,
        "tags_exist": StructureTag.objects.exists(),
        "data_tables_page_length": STRUCTURES_DEFAULT_PAGE_LENGTH,
        "data_tables_paging": STRUCTURES_PAGING_ENABLED,
        "show_jump_gates_tab": STRUCTURES_SHOW_JUMP_GATES,
        "last_updated": Owner.objects.structures_last_updated(),
    }
    return render(request, "structures/main.html", context)


@login_required
@permission_required("structures.basic_access")
def structure_list_data(request) -> JsonResponse:
    """returns structure list in JSON for AJAX call in structure_list view"""
    tags_raw = request.GET.get(QUERY_PARAM_TAGS)
    tags = tags_raw.split(",") if tags_raw else None
    structures = Structure.objects.visible_for_user(request.user, tags)
    serializer = StructureListSerializer(queryset=structures, request=request)
    return JsonResponse({"data": serializer.to_list()})


@login_required
@permission_required("structures.view_structure_fit")
def structure_details(request, structure_id):
    """Main view of the structure fit"""

    class Slot(IntEnum):
        HIGH = 14
        MEDIUM = 13
        LOW = 12
        RIG = 1137
        SERVICE = 2056

        def image_url(self) -> str:
            """Return url to image file for this slot variant"""
            id_map = {
                self.HIGH: "h",
                self.MEDIUM: "m",
                self.LOW: "l",
                self.RIG: "r",
                self.SERVICE: "s",
            }
            try:
                slot_num = type_attributes[self.value]
                return staticfiles_storage.url(
                    f"structures/img/pannel/{slot_num}{id_map[self.value]}.png"
                )
            except KeyError:
                return ""

    class FakeEveType:
        def __init__(self, id, name):
            self.id = id
            self.name = name
            self.profile_url = ""

        def icon_url(self, size=64) -> str:
            return eveimageserver.type_icon_url(self.id, size)

    class FakeAsset:
        """Fake asset object for showing additional information in the asset list."""

        def __init__(self, name, quantity, eve_type_id):
            self.name = name
            self.quantity = quantity
            self.eve_type_id = eve_type_id
            self.eve_type = FakeEveType(eve_type_id, name)
            self.is_singleton = False

    def extract_slot_assets(fittings: list, slot_name: str) -> list:
        """Return assets for slot sorted by slot number"""
        return [
            asset[0]
            for asset in sorted(
                [
                    (asset, asset.location_flag[-1])
                    for asset in fittings
                    if asset.location_flag.startswith(slot_name)
                ],
                key=lambda x: x[1],
            )
        ]

    def patch_fighter_tube_quantities(fighter_tubes):
        eve_type_ids = {item.eve_type_id for item in fighter_tubes}
        eve_types = [
            EveType2.objects.get_or_create_esi(
                id=eve_type_id, enabled_sections=[EveType2.Section.DOGMAS]
            )[0]
            for eve_type_id in eve_type_ids
        ]
        squadron_sizes = {
            eve_type.id: int(
                eve_type.dogma_attributes.get(
                    eve_dogma_attribute=EveAttributeId.SQUADRON_SIZE.value
                ).value
            )
            for eve_type in eve_types
        }
        for item in fighter_tubes:
            try:
                squadron_size = squadron_sizes[item.eve_type_id]
            except KeyError:
                pass
            else:
                item.quantity = squadron_size
                item.is_singleton = False

    structure = get_object_or_404(
        Structure.objects.select_related(
            "owner",
            "owner__corporation",
            "owner__corporation__alliance",
            "eve_type",
            "eve_type__eve_group",
            "eve_solar_system",
            "eve_solar_system__eve_constellation",
            "eve_solar_system__eve_constellation__eve_region",
        ),
        id=structure_id,
    )
    type_attributes = {
        obj["eve_dogma_attribute_id"]: int(obj["value"])
        for obj in EveTypeDogmaAttribute.objects.filter(
            eve_type_id=structure.eve_type_id
        ).values("eve_dogma_attribute_id", "value")
    }
    slot_image_urls = {
        "high": Slot.HIGH.image_url(),
        "med": Slot.MEDIUM.image_url(),
        "low": Slot.LOW.image_url(),
        "rig": Slot.RIG.image_url(),
        "service": Slot.SERVICE.image_url(),
    }
    assets = structure.items.select_related("eve_type")
    high_slots = extract_slot_assets(assets, "HiSlot")
    med_slots = extract_slot_assets(assets, "MedSlot")
    low_slots = extract_slot_assets(assets, "LoSlot")
    rig_slots = extract_slot_assets(assets, "RigSlot")
    service_slots = extract_slot_assets(assets, "ServiceSlot")
    fighter_tubes = extract_slot_assets(assets, "FighterTube")
    patch_fighter_tube_quantities(fighter_tubes)
    assets_grouped = {"ammo_hold": [], "fighter_bay": [], "fuel_bay": []}
    for asset in assets:
        if asset.location_flag == StructureItem.LocationFlag.CARGO:
            assets_grouped["ammo_hold"].append(asset)
        elif asset.location_flag == StructureItem.LocationFlag.FIGHTER_BAY:
            assets_grouped["fighter_bay"].append(asset)
        elif asset.location_flag == StructureItem.LocationFlag.STRUCTURE_FUEL:
            assets_grouped["fuel_bay"].append(asset)
        else:
            assets_grouped[asset.location_flag] = asset
    if structure.is_upwell_structure:
        assets_grouped["fuel_usage"] = [
            FakeAsset(
                name="Fuel blocks per day (est.)",
                quantity=structure.structure_fuel_usage(),
                eve_type_id=24756,
            )
        ]
    modules_count = len(high_slots + med_slots + low_slots + rig_slots + service_slots)
    fuel_blocks_total = (
        functools.reduce(
            lambda x, y: x + y, [obj.quantity for obj in assets_grouped["fuel_bay"]]
        )
        if assets_grouped["fuel_bay"]
        else 0
    )
    ammo_total = (
        functools.reduce(
            lambda x, y: x + y, [obj.quantity for obj in assets_grouped["ammo_hold"]]
        )
        if assets_grouped["ammo_hold"]
        else 0
    )
    fighters_consolidated = assets_grouped["fighter_bay"] + fighter_tubes
    fighters_total = (
        functools.reduce(
            lambda x, y: x + y, [obj.quantity for obj in fighters_consolidated]
        )
        if fighters_consolidated
        else 0
    )
    context = {
        "fitting": assets,
        "slots": slot_image_urls,
        "slot_assets": {
            "high_slots": high_slots,
            "med_slots": med_slots,
            "low_slots": low_slots,
            "rig_slots": rig_slots,
            "service_slots": service_slots,
            "fighter_tubes": fighter_tubes,
        },
        "assets_grouped": assets_grouped,
        "structure": structure,
        "modules_count": modules_count,
        "fuel_blocks_total": fuel_blocks_total,
        "fighters_total": fighters_total,
        "ammo_total": ammo_total,
        "last_updated": structure.owner.assets_last_update_at,
    }
    return render(request, "structures/modals/structure_details.html", context)


@login_required
@permission_required("structures.basic_access")
def poco_details(request, structure_id):
    """Shows details modal for a POCO."""

    structure = get_object_or_404(
        Structure.objects.select_related(
            "owner",
            "eve_type",
            "eve_solar_system",
            "eve_solar_system__eve_constellation",
            "eve_solar_system__eve_constellation__eve_region",
            "poco_details",
            "eve_planet",
        ).filter(eve_type=EveTypeId.CUSTOMS_OFFICE, poco_details__isnull=False),
        id=structure_id,
    )
    context = {
        "structure": structure,
        "details": structure.poco_details,
        "last_updated": structure.last_updated_at,
    }
    return render(request, "structures/modals/poco_details.html", context)


@login_required
@permission_required("structures.basic_access")
def starbase_detail(request, structure_id):
    """Shows detail modal for a starbase."""

    structure = get_object_or_404(
        Structure.objects.select_related(
            "owner",
            "owner__corporation",
            "owner__corporation__alliance",
            "eve_type",
            "eve_type__eve_group",
            "eve_solar_system",
            "eve_solar_system__eve_constellation",
            "eve_solar_system__eve_constellation__eve_region",
            "starbase_detail",
            "eve_moon",
        ).filter(starbase_detail__isnull=False),
        id=structure_id,
    )
    fuels = structure.starbase_detail.fuels.select_related("eve_type").order_by(
        "eve_type__name"
    )
    assets = defaultdict(int)
    for item in structure.items.select_related("eve_type"):
        assets[item.eve_type_id] += item.quantity
    eve_types = EveType.objects.in_bulk(id_list=assets.keys())
    modules = sorted(
        [
            {"eve_type": eve_types.get(eve_type_id), "quantity": quantity}
            for eve_type_id, quantity in assets.items()
        ],
        key=lambda obj: obj["eve_type"].name,
    )
    modules_count = (
        functools.reduce(lambda x, y: x + y, [obj["quantity"] for obj in modules])
        if modules
        else 0
    )
    try:
        fuel_blocks_count = (
            structure.starbase_detail.fuels.filter(
                eve_type__eve_group_id=EveGroupId.FUEL_BLOCK
            )
            .first()
            .quantity
        )
    except AttributeError:
        fuel_blocks_count = None
    context = {
        "structure": structure,
        "detail": structure.starbase_detail,
        "fuels": fuels,
        "modules": modules,
        "modules_count": modules_count,
        "fuel_blocks_count": fuel_blocks_count,
        "last_updated_at": structure.last_updated_at,
    }
    return render(request, "structures/modals/starbase_detail.html", context)


@login_required
@permission_required("structures.add_structure_owner")
@token_required(scopes=Owner.get_esi_scopes())
def add_structure_owner(request, token):
    token_char = get_object_or_404(EveCharacter, character_id=token.character_id)
    try:
        character_ownership = CharacterOwnership.objects.get(
            user=request.user, character=token_char
        )
    except CharacterOwnership.DoesNotExist:
        character_ownership = None
        messages_plus.error(
            request,
            format_html(
                _(
                    "You can only use your main or alt characters "
                    "to add corporations. "
                    "However, character %s is neither. "
                )
                % format_html("<strong>{}</strong>", token_char.character_name)
            ),
        )
        return redirect("structures:index")
    try:
        corporation = EveCorporationInfo.objects.get(
            corporation_id=token_char.corporation_id
        )
    except EveCorporationInfo.DoesNotExist:
        corporation = EveCorporationInfo.objects.create_corporation(
            token_char.corporation_id
        )
    owner, created = Owner.objects.update_or_create(
        corporation=corporation, defaults={"is_active": True}
    )
    owner.add_character(character_ownership)
    if created:
        default_webhooks = Webhook.objects.filter(is_default=True)
        if default_webhooks:
            for webhook in default_webhooks:
                owner.webhooks.add(webhook)
            owner.save()

    if owner.characters.count() == 1:
        tasks.update_all_for_owner.delay(owner_pk=owner.pk, user_pk=request.user.pk)
        messages_plus.info(
            request,
            format_html(
                _(
                    "%(corporation)s has been added with %(character)s "
                    "as sync character. "
                    "We have started fetching structures and notifications "
                    "for this corporation and you will receive a report once "
                    "the process is finished."
                )
                % {
                    "corporation": format_html("<strong>{}</strong>", owner),
                    "character": format_html("<strong>{}</strong>", token_char),
                }
            ),
        )
        if STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED:
            with translation.override(STRUCTURES_DEFAULT_LANGUAGE):
                notify_admins(
                    message=_(
                        "%(corporation)s was added as new "
                        "structure owner by %(user)s."
                    )
                    % {"corporation": owner, "user": request.user.username},
                    title=_("%s: Structure owner added: %s") % (__title__, owner),
                )
    else:
        messages_plus.info(
            request,
            format_html(
                _(
                    "%(character)s has been added to %(corporation)s "
                    "as sync character. "
                    "You now have %(characters_count)d sync character(s) configured."
                )
                % {
                    "corporation": format_html("<strong>{}</strong>", owner),
                    "character": format_html("<strong>{}</strong>", token_char),
                    "characters_count": owner.characters_count(),
                }
            ),
        )
        if STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED:
            with translation.override(STRUCTURES_DEFAULT_LANGUAGE):
                notify_admins(
                    message=_(
                        "%(character)s was added as sync character to "
                        "%(corporation)s by %(user)s.\n"
                        "We now have %(characters_count)d sync character(s) configured."
                    )
                    % {
                        "character": token_char,
                        "corporation": owner,
                        "user": request.user.username,
                        "characters_count": owner.characters_count(),
                    },
                    title=_("%s: Character added to: %s") % (__title__, owner),
                )
    return redirect("structures:index")


def service_status(request):
    """Public view to 3rd party monitoring.

    This is view allows running a 3rd party monitoring on the status
    of this services. Service will be reported as down if any of the
    configured structure or notifications syncs fails or is delayed
    """
    status_ok = True
    for owner in Owner.objects.filter(is_included_in_service_status=True):
        status_ok = status_ok and owner.are_all_syncs_ok

    if status_ok:
        return HttpResponse(_("service is up"))
    else:
        return HttpResponseServerError(_("service is down"))


def poco_list_data(request) -> JsonResponse:
    """List of public POCOs for DataTables."""
    pocos = Structure.objects.filter(
        eve_type__eve_group__eve_category_id=EveCategoryId.ORBITAL,
        owner__are_pocos_public=True,
    )
    serializer = PocoListSerializer(queryset=pocos, request=request)
    return JsonResponse({"data": serializer.to_list()})


@login_required
@permission_required("structures.basic_access")
def structure_summary_data(request) -> JsonResponse:
    summary_qs = (
        Structure.objects.values(
            "owner__corporation__corporation_id",
            "owner__corporation__corporation_name",
            "owner__corporation__alliance__alliance_name",
        )
        .annotate(
            ec_count=Count(
                "id", filter=Q(eve_type__eve_group=EveGroupId.ENGINEERING_COMPLEX)
            )
        )
        .annotate(
            refinery_count=Count(
                "id", filter=Q(eve_type__eve_group=EveGroupId.REFINERY)
            )
        )
        .annotate(
            citadel_count=Count("id", filter=Q(eve_type__eve_group=EveGroupId.CITADEL))
        )
        .annotate(
            upwell_count=Count(
                "id",
                filter=Q(eve_type__eve_group__eve_category=EveCategoryId.STRUCTURE),
            )
        )
        .annotate(poco_count=Count("id", filter=Q(eve_type=EveTypeId.CUSTOMS_OFFICE)))
        .annotate(
            starbase_count=Count(
                "id", filter=Q(eve_type__eve_group__eve_category=EveCategoryId.STARBASE)
            )
        )
    )
    data = list()
    for row in summary_qs:
        other_count = (
            row["upwell_count"]
            - row["ec_count"]
            - row["refinery_count"]
            - row["citadel_count"]
        )
        total = row["upwell_count"] + row["poco_count"] + row["starbase_count"]
        corporation_icon_url = eveimageserver.corporation_logo_url(
            row["owner__corporation__corporation_id"], size=64
        )
        corporation_icon = image_html(corporation_icon_url, size=32)
        alliance_name = default_if_none(
            row["owner__corporation__alliance__alliance_name"], ""
        )
        data.append(
            {
                "id": int(row["owner__corporation__corporation_id"]),
                "corporation_icon": corporation_icon,
                "corporation_name": row["owner__corporation__corporation_name"],
                "alliance_name": alliance_name,
                "citadel_count": row["citadel_count"],
                "ec_count": row["ec_count"],
                "refinery_count": row["refinery_count"],
                "other_count": other_count,
                "poco_count": row["poco_count"],
                "starbase_count": row["starbase_count"],
                "total": total,
            }
        )
    return JsonResponse({"data": data})


def jump_gates_list_data(request) -> JsonResponse:
    """List of jump gates for DataTables."""
    jump_gates = Structure.objects.visible_for_user(request.user).filter(
        eve_type_id=EveTypeId.JUMP_GATE
    )
    serializer = JumpGatesListSerializer(queryset=jump_gates)
    return JsonResponse({"data": serializer.to_list()})



# Generated by Django 2.2.11 on 2020-04-13 20:02

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0009_localizations"),
    ]

    operations = [
        migrations.CreateModel(
            name="EveSovereigntyMap",
            fields=[
                (
                    "solar_system_id",
                    models.PositiveIntegerField(primary_key=True, serialize=False),
                ),
                (
                    "alliance_id",
                    models.PositiveIntegerField(
                        blank=True,
                        db_index=True,
                        help_text="alliance who holds sov for this system",
                        null=True,
                    ),
                ),
                (
                    "corporation_id",
                    models.PositiveIntegerField(
                        blank=True,
                        db_index=True,
                        help_text="corporation who holds sov for this system",
                        null=True,
                    ),
                ),
                (
                    "faction_id",
                    models.PositiveIntegerField(
                        blank=True,
                        db_index=True,
                        help_text="faction who holds sov for this system",
                        null=True,
                    ),
                ),
                (
                    "last_updated",
                    models.DateTimeField(
                        blank=True,
                        db_index=True,
                        default=None,
                        help_text="When this object was last updated from ESI",
                        null=True,
                    ),
                ),
            ],
        ),
        migrations.AlterModelOptions(
            name="structuretag",
            options={"ordering": ["order", "name"]},
        ),
        migrations.AddField(
            model_name="structuretag",
            name="is_user_managed",
            field=models.BooleanField(
                default=True,
                help_text="if False this tag is created and managed by the system and can not be modified by users",
            ),
        ),
        migrations.AddField(
            model_name="structuretag",
            name="order",
            field=models.PositiveIntegerField(
                blank=True,
                default=100,
                help_text="number defining the order tags are shown. custom tags can not have an order below 100",
                validators=[django.core.validators.MinValueValidator(100)],
            ),
        ),
        migrations.AlterField(
            model_name="structuretag",
            name="is_default",
            field=models.BooleanField(
                default=False,
                help_text="if true this custom tag will automatically be added to new structures",
            ),
        ),
        migrations.AlterField(
            model_name="structuretag",
            name="style",
            field=models.CharField(
                blank=True,
                choices=[
                    ("default", "grey"),
                    ("primary", "dark blue"),
                    ("success", "green"),
                    ("info", "light blue"),
                    ("warning", "orange"),
                    ("danger", "red"),
                ],
                default="default",
                help_text="color style of tag",
                max_length=16,
            ),
        ),
    ]
